"""
Module containing routines for computation of observables.
"""
 
from tightbinder.result import Spectrum
from tightbinder.system import System
from tightbinder.models import SlaterKoster

import numpy as np
import scipy.sparse as sp
import matplotlib.pyplot as plt
from matplotlib.axes import Axes
from typing import Union, List, Tuple
from copy import deepcopy


def _retarded_green_function(w: float, e: float, delta: float) -> complex:
    """ 
    Routine to compute the retarded Green function, to be used to obtain
    the DOS.

    :param w: Value of frequency.
    :param e: Value of energy.
    :param delta: Value of broadening.
    :return: Retarded Green's function evaluated at w, e. 
    """
    
    green = 1./(w - e + 1j*delta)
    return green


def dos(result: Spectrum, energy: float = None, delta: float = 0.1, npoints: int = 200) -> Tuple[List[float], List[float]]:
    """ 
    Routine to compute the density of states from a Result object.

    :param result: Result object.
    :param energy: Value of energy where DoS is computed. If None, then the DoS is computed
        for the whole energy window of the spectrum.
    :param delta: Delta broadening, defaults to 0.01. 
    :param npoints: Number of energy points to use if sampling whole DoS.
    :return: Returns duple with values of dos, and energies where was computed.
    """
    
    nkpoints = result.eigen_energy.shape[1]
    eigval = result.eigen_energy.reshape(-1, )
    if not energy:
        emin, emax = np.min(result.eigen_energy)*1.5, np.max(result.eigen_energy)*1.5
        energies = np.linspace(emin, emax, npoints, endpoint=True)
    else:
        energies = [energy]

    energies = np.array(energies)
    dos = []
    for energy in energies:
        density = (np.sum(-np.imag(_retarded_green_function(energy, eigval, delta))/np.pi)
                /(nkpoints))
        dos.append(density)

    return dos, energies


def dos_kpm(system: System, energy: float = None, npoints: int = 200, nmoments: int = 30, r: int = 10) -> Tuple[List[float], List[float]]:
    """ 
    Routine to compute the density of states using the Kernel Polynomial Method.
    Intended to be used with supercells and k=0.

    :param system: System whose density of states we want to compute.
    :param energy: Value of energy where we want to obtain the DoS. If None, DoS is computed for all
        energy window of spectrum.
    :param npoints: Number of energy points to consider if computing full DoS. Defaults to 200.
    :param nmoments: Number of moments to use in the KPM expansion. Defaults to 30.
    :param r: Number of samples to use in the stochastic evaluation of the KPM traces.
        Defaults to 10.
    :return:
    """

    if system.matrix_type != "sparse":
        print("Warning: KPM computations are intended to be run with sparse matrices (faster and less memory usage)")

    print("Computing DOS using the KPM...")
    h = system.hamiltonian_k([0., 0., 0.])
    if system.matrix_type == "dense":
        h = sp.bsr_matrix(h)
    # h spectrum has to be between -1 and 1
    h_norm = sp.linalg.norm(h, ord=np.inf)
    h /= h_norm
    
    if not energy:
        emin, emax = -1, 1
        energies = np.linspace(emin, emax, npoints, endpoint=True)[1:-1]
    else:
        energies = [energy/h_norm]

    energies = np.array(energies)
    moments = compute_dos_momentum(nmoments, h, r)
    jackson = jackson_kernel(nmoments)
    moments = moments * jackson
    densities = []
    for energy in energies:
        polynomials = chebyshev_polynomial_values(nmoments, energy)
        density = moments[0] * polynomials[0] + 2*np.sum(moments[1:] * polynomials[1:])
        density = density/(np.pi*np.sqrt(1 - energy**2))
        densities.append(density)

    energies = energies * h_norm
    densities = np.array(densities).real * h_norm
    return densities, energies


def jackson_kernel(nmoments: int) -> List[float]:
    """ 
    Routine to calculate the Jackson kernel to improve the KPM calculations. 
    
    :param nmoments: Number of moments of KPM expansions.
    :return: Jackson kernel for n=0, ..., nmoments - 1
    """

    g = [((nmoments - n + 1)*np.cos(np.pi*n/(nmoments + 1)) +
          np.sin(np.pi*n/(nmoments + 1))/np.tan(np.pi*n/(nmoments + 1)))/(nmoments + 1) for n in range(1, nmoments)]
    g = [1] + g  # n = 0 gives NaN instead of 1
    
    return g


def chebyshev_polynomial_values(n: int, e: float) -> np.ndarray:
    """ 
    Chebyshev polynomial of order n on value e. 
    
    :param n: Maximum order of Chebyshev polynomial.
    :param e: Value where the polynomials is evaluated (-1 < e < 1).
    :return: List with all polynomials from order 0 to n - 1 evaluated on e.
    """
    
    polynomials = [1, e]
    if n >= 2:
        for i in range(2, n):
            polynomials.append(2*e*polynomials[i - 1] - polynomials[i - 2]) 
    return np.array(polynomials)


def compute_dos_momentum(n: int, h: np.ndarray, r: int) -> np.ndarray:
    """ 
    Routine to compute up to the n-th momentum associated to a Chebyshev polynomial expansion
    of the density of states. Instead of computing the whole trace of h, we perform the stochastic
    evaluation of the trace over a small set of random vectors. 
    
    :param n: Total number of momentum computed for DoS calculation.
    :param h: Hamiltonian, usually on k=0 (intended for supercell calculations).
    :param r: Number of samples on stochastic ealuation of trace.
    :return: Array with all momentum.
    """

    d = h.shape[0]

    # Generate sample of random vectors and normalize them
    vectors = np.random.uniform(-1, 1, (r, d))
    norm = np.sqrt(np.diag(np.dot(vectors, vectors.T)))
    vectors = vectors / norm[:, None]
    vectors = vectors.T

    # Evaluate each momentum
    momentum = []
    operators = [sp.eye(h.shape[0], h.shape[1]).dot(vectors), h.dot(vectors)]
    momentum.append(np.sum(vectors.T.dot(operators[0]).diagonal())/(r*d))
    momentum.append(np.sum(vectors.T.dot(operators[1]).diagonal())/(r*d))
    for i in range(2, n):
        operator = 2*h.dot(operators[1]) - operators[0]
        mu = np.sum(vectors.T.dot(operator).diagonal())/(r*d)
        momentum.append(mu)
        operators[0] = operators[1]
        operators[1] = operator

    return np.array(momentum)


def plot_dos(result: Spectrum, npoints: int = 100, delta: float = 0.3, ax: Axes = None) -> None:
    """ 
    Routine to plot the density of states using the retarded Green function. 
    
    :param result: Spectrum object containing the eigenenergies.
    :param npoints: Number of points used to sample the DoS. Defaults to 100.
    :param delta: Broadening of the states. Defaults to 0.3.
    :param ax: Axes object to plot the DoS in it. Defaults to None.
    """

    if not ax:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    density, energies = dos(result, delta=delta, npoints=npoints)

    ax.plot(energies, density)
    ax.set_xlabel(r'E (eV)')
    ax.set_ylabel(r'DOS')

    return density, energies


def plot_dos_kpm(system: System, npoints: int = 200, nmoments: int = 100, r: int = 10, ax: Axes = None) -> None:
    """
    Routine to plot the density of states using the KPM. 
    
    :param system: System whose DoS we want to obtain and plot.
    :param npoints: Number of points used to sample the DoS. Defaults to 200.
    :param nmoments: Number of moments in KPM expansion. Defaults to 100.
    :param r: Number of samples for stochastic evaluation of trace in KPM. Defaults to 100.
    :param ax: Optional Axes object to draw the plot in.
    """

    if not ax:
        fig = plt.figure()
        ax = fig.add_subplot(111)

    dos, energies = dos_kpm(system, npoints=npoints, nmoments=nmoments, r=r)

    ax.plot(energies, dos)
    ax.set_title('DOS (KPM)')
    ax.set_xlabel(r'E (eV)')
    ax.set_ylabel(r'DOS')

    return dos, energies


def fermi_energy(dos: list, energy: list, system: System) -> float:
    """ 
    Routine to compute the Fermi energy from the density of states
    integrating up to the number of electrons in the system. It is based on 
    trapezoidal integration of the DoS.
    
    :param dos: List with density of states values.
    :param energy: List with energies where the DoS are evaluated.
    :param system: System object (used to provide filling and basisdim).
    :return: Fermi energy.
    """

    # First normallize to basisdim (maximum number of states)
    area = np.trapz(dos, energy)
    dos = dos / area * system.basisdim
    # Then search
    for i in range(len(dos)):
        nstates = np.trapz(dos[:i], energy[:i])
        if nstates > system.filling:
            break
    
    return energy[i - 1]


def compute_projector_momentum(nmoments: int, energy: int) -> np.ndarray:
    """ 
    Routine to compute the momentum from the KPM expansion of the ground state projector. 
    
    :param nmoments: Number of momentum we want to compute.
    :param energy: Energy up to which states are considered.
    :return: List of moments.
    """

    first_moment = [1 - np.arccos(energy)/np.pi]
    moments = [-2*np.sin(m*np.arccos(energy))/(m*np.pi) for m in np.arange(1, nmoments)]
    moments = first_moment + moments

    return np.array(moments)

def ground_state_projector(nmoments: int, h: np.ndarray, energy: int) -> np.ndarray:
    """ 
    Routine to compute the ground state projector using the KPM. Note that this calculation
    is done using sparse arrays, since it results in faster computation.
    
    :param nmoments: Number of moments in the expansion.
    :param h: Bloch Hamiltonian, usually at k=0 for supercells.
    :param energy: Energy up to which the states are considered. To compute the ground state
        projector, this should be the Fermi energy.
    :return: Projector.
    """

    h_norm = np.linalg.norm(h)
    h /= h_norm
    h = sp.bsr_matrix(h)
    print("Cast to sparse done")
    energy /= h_norm

    print("Computing moments...")
    moments = compute_projector_momentum(nmoments, energy)
    jackson = jackson_kernel(nmoments)
    moments = moments * jackson
    operators = [sp.eye(h.shape[0], h.shape[1], format='csr'), h]
    projector = moments[0]*operators[0] + moments[1]*operators[1]
    for i in range(2, nmoments):
        print(f"Moment: {i}")
        operator = 2*h.dot(operators[1]) - operators[0]
        projector += moments[i] * operator
        operators[0] = operators[1]
        operators[1] = operator

    return projector.toarray()


def restricted_density_matrix(system: System, partition: list, nmoments: int = 100, npoints: int = 200, 
                              nmoments_dos: int = 300, r: int = 10) -> np.ndarray:
    """ 
    Routine to compute the one-particle density matrix, restricted to one spatial partition of the system. First it estimates the Fermi
    energy from the DOS computed with the KPM, and then uses it to compute the reduced density matrix. 
    
    :param system: System whose reduced density matrix we want to compute.
    :param partition: List of indices of atoms where the reduced density matrix is computed.
    :param nmoments: Number of moments in the KPM expansion. Defaults to 100.
    :param npoints: Number of used to sample the density of states. Defaults to 200.
    :param nmoments_dos: Number of moments in the KPM expansion of the DoS. Defaults to 300.
    :param r: Number of samples in the stochastic evaluation of the trace. Defaults to 10.
    :return: Reduced density matrix.
    """

    if system.matrix_type != "sparse":
        print("Warning: KPM computations are intended to be run with sparse matrices (faster and less memory usage)")
    
    h = system.hamiltonian_k([[0., 0., 0.]])
    if system.matrix_type == "dense":
        h = sp.bsr_matrix(h)

    h_norm = sp.linalg.norm(h)
    h /= h_norm
    
    # First compute fermi energy from dos
    dos, energy = dos_kpm(system, npoints=npoints, nmoments=nmoments_dos, r=r)
    efermi = fermi_energy(dos, energy, system)/h_norm

    orbitals = []
    counter = 0
    for n in range(system.natoms):
        norb = system.norbitals[int(system.motif[n, 3])]
        if n in partition:
            for _ in range(0, norb):
                orbitals.append(counter)
                counter += 1
        else:
            counter += norb
    
    orbital_cols = np.arange(len(orbitals))
    orbital_values = np.ones(len(orbitals))
    orbital_matrix = sp.csr_matrix((orbital_values, (orbitals, orbital_cols)), shape=(system.basisdim, len(orbitals)))

    moments = compute_projector_momentum(nmoments, efermi)
    jackson = jackson_kernel(nmoments)
    moments = moments * jackson
    operators = [orbital_matrix, h.dot(orbital_matrix)]

    density_matrix = moments[0]*operators[0] + moments[1]*operators[1]
    for i in range(2, nmoments):
        print(f"Moment: {i}")
        operator = 2*h.dot(operators[1]) - operators[0]
        density_matrix += moments[i] * operator
        operators[0] = operators[1]
        operators[1] = operator

    density_matrix = orbital_matrix.transpose().dot(density_matrix).toarray()

    print(f"Fermi energy: {efermi}")
    return density_matrix


class TransportDevice:
    def __init__(self, system: System, left_lead: Union[List, np.ndarray], right_lead: Union[List, np.ndarray],
                 period: Union[List, float], mode: str = "default") -> None:
        """
        Constructor of TransportDevice class to model a transport setup.

        :param system: System to compute transmission
        :param left_lead: Unit cell of left lead. Array or list where each row contains the position
            and chemical species of each atom of the lead.
        :param right_lead: Same as left lead.
        :param period: Distance between consecutive lead unit cells. If given one value, it is used
            for both right and left leads. If given a list with two values, the first one specifies left
            period, and second one the right period.
        :param mode: Either "default" or "direct". Default mode uses the SK configuration to establish the
            system-leads bonds, the intra-lead and the lead-lead bonds. Intended to be used with crystalline
            or quasicrystalline situations (neighbour based search).
            Direct mode instead connects the leads to the system straight: each atom of the lead has only
            one bond (the first found) to the system. The intra-lead and lead-lead bonds are determined 
            using the SK configuration. Intended to be used with amorphous systems (radius based search).
        """
         
        
        if not left_lead.size or not right_lead.size:
            raise ValueError("Must provide a non-empty list for leads")

        if type(period) != list:
            period = [period, period]
        else:
            if len(period) != 2:
                raise ValueError("period can hold only two values, [left_period, right_period]")
        
        if mode not in ["default", "direct"]:
            raise ValueError("Invalid mode. Must be either 'default' or 'direct'.")
        
        self.system = system
        self.left_lead = left_lead
        self.right_lead = right_lead
        self.period = period
        self.mode = mode
        self.bonds = []
        
        device, left_lead, right_lead = self.__attach_leads()

        self.device_h = device
        self.left_lead_h = left_lead
        self.right_lead_h = right_lead


    def transmission(self, minE: float, maxE: float, npoints: int = 100, 
                    delta: float = 1E-7, method: str = "LS") -> Tuple[List, List]:
        """
        Function to compute the transmission function T(E) of a system. One has to specify the unit
        cell of each lead in terms of the positions and chemical species. Only two terminal setups are allowed. 
        The lead-system coupling and the lead-lead coupling are computed using the corresponding Slater-Koster amplitudes.

        :param minE: Minimum value of energy window where the transmission is computed.
        :param maxE: Maximum value of energy window.
        :param npoints: Sampling of energy window. Defaults to 100.
        :param delta: Value of broadening used in Green's functions. Defaults to 1E-7.
        :param method: Method to convergence the self-energies of the leads. Options are 'LS' (Lopez Sancho method)
            and 'LM' (Linear Mixing). Defaults to 'LS'.
        :return: List of transmission values and energy window.
        """


        device_dim = self.device_h.shape[0]
        energies = np.linspace(minE, maxE, npoints)
        currents = []
        extended_lead_selfenergy_L = sp.lil_matrix((device_dim, device_dim), dtype=np.complex_)
        extended_lead_selfenergy_R = sp.lil_matrix((device_dim, device_dim), dtype=np.complex_)

        for energy in energies:
            lead_selfenergy_L = self.__lead_selfenergy(energy, self.left_lead_h[0], self.left_lead_h[1].transpose().conjugate(), delta, method=method)  
            coupling_L = 1j*(lead_selfenergy_L - lead_selfenergy_L.transpose().conjugate())   
            extended_lead_selfenergy_L[:lead_selfenergy_L.shape[0], :lead_selfenergy_L.shape[0]] = coupling_L

            lead_selfenergy_R = self.__lead_selfenergy(energy, self.right_lead_h[0], self.right_lead_h[1], delta, method=method)
            coupling_R = 1j*(lead_selfenergy_R - lead_selfenergy_R.transpose().conjugate())
            extended_lead_selfenergy_R[(device_dim - lead_selfenergy_R.shape[0]):, (device_dim - lead_selfenergy_R.shape[0]):] = coupling_R

            device_green = self.__device_green_function(energy, self.device_h, lead_selfenergy_L, lead_selfenergy_R, delta)

            T = (extended_lead_selfenergy_L @ device_green.transpose().conjugate() @ extended_lead_selfenergy_R @ device_green).trace()
            currents.append(T.real)

        return (currents, energies)


    def conductance(self, delta: float = 1E-7, method: str = "LS") -> float:
        """
        Routine to compute the conductance at zero temperature in equilibrium.

        :param delta: Value of broadening used in Green's functions. Defaults to 1E-7.
        :param method: Method to convergence the self-energies of the leads. Options are 'LS' (Lopez Sancho method)
            and 'LM' (Linear Mixing). Defaults to 'LS'.
        :return: Value of conductance.
        """

        self.system.initialize_hamiltonian()
        spectrum = self.system.solve()
        fermi_energy = spectrum.calculate_fermi_energy(self.system.filling)
        G = self.transmission(fermi_energy, fermi_energy, 1, delta, method)

        return G[0][0]
            

    def __find_direct_bonds(self, first_group: Union[List, np.ndarray], second_group: Union[List, np.ndarray], r: float) -> List:
        """
        Private method to determine the bonds from a first group of atoms to a second. 
        For each atom of the first group, it determines only the first bond to an atom of the
        second group. To be used by the direct mode of the transmission routine.

        :param first_group: List of atomic positions.
        :param second_group: List of atomic positions.
        :param r: Cutoff for bond search.
        :return: List of direct bonds from first group to second group of atoms.
        """

        first_group = np.array(first_group)
        second_group = np.array(second_group)
        if self.system.bravais_lattice is not None:
            cells = [self.system.bravais_lattice[0], np.array([0., 0., 0.]), -self.system.bravais_lattice[0]]
        else:
            cells = [np.array([0., 0., 0.])]
        bonds = []
        it = 0
        for i, atom in enumerate(first_group):
            for cell in cells:
                final_atoms = np.copy(second_group)[:, :3] + cell - atom[:3] 
                distances = np.linalg.norm(final_atoms, axis=1)
                indices = np.argsort(distances)
                for index in indices:
                    if distances[index] < r:
                        bonds.append([i, index, cell, "1"])
        
        return bonds


    def __attach_leads(self) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Private method to extend the Hamiltonian to include a part of the leads. It assumes that the system has
        rectangle shape. 

        :param system: System to attach leads to.
        :param left_leaf: List of positions and chemical species of atoms of the left lead.
        :param right_boundary: List of positions and chemical species of atoms of the right lead.
        :param period: Distance between consecutive lead unit cells. If given one value, it is used
            for both right and left leads. If given a list with two values, the first one specifies left
            period, and second one the right period.
        :param mode: Lead-system connection mode. Either "default" or "direct".
        :return: (Hd,[Hl, Vl],[Hr, Vr]) Hd = Hamiltonian of device with leads attached.
            [Hl, Vl] = Left lead hamiltonian and coupling. [Hr, Vr] = Right lead hamiltonian and coupling.
        """


        max_nn = max(int(value) for value in self.system.configuration["SK amplitudes"].keys())
        device_motif = np.concatenate((self.left_lead, self.system.motif, self.right_lead))
        model_device = deepcopy(self.system)
        model_device.matrix_type = "sparse"
        model_device.motif = device_motif
        if self.mode == "default":
            model_device.initialize_hamiltonian()
        else:
             # Bonds within left lead unit cell
            left_lead = deepcopy(self.system)
            left_lead.motif = self.left_lead
            left_lead.find_neighbours(mode="minimal", nn=max_nn)
            fnn = left_lead.compute_first_neighbour_distance()

            # First obtain bonds between left lead and system
            disp = len(self.left_lead)
            left_bonds = self.__find_direct_bonds(self.left_lead, self.system.motif, fnn * 1.3)
            left_bonds = [[i, j + disp, cell, nn] for [i, j, cell, nn] in left_bonds]
            reversed_bonds = [[j, i, -cell, nn] for [i, j, cell, nn] in left_bonds]
            left_bonds = left_bonds + reversed_bonds + left_lead.bonds

            # Bonds within right lead
            right_lead = deepcopy(self.system)
            right_lead.motif = self.right_lead
            right_lead.find_neighbours(mode="minimal", nn=max_nn)
            fnn = right_lead.compute_first_neighbour_distance()

            # Obtain bonds between right lead and system
            disp = len(self.left_lead) + len(self.system.motif)
            right_bonds = self.__find_direct_bonds(self.right_lead, self.system.motif, fnn * 1.3)
            right_bonds = [[i + disp, j + len(self.left_lead), cell, nn] for [i, j, cell, nn] in right_bonds]
            reversed_bonds = [[j, i, -cell, nn] for [i, j, cell, nn] in right_bonds]
            right_bonds = right_bonds + reversed_bonds + [[i + disp, j + disp, cell, nn] for [i, j, cell, nn] in right_lead.bonds]

            self.system.initialize_hamiltonian()
            system_bonds = [[i + len(self.left_lead), j + len(self.left_lead), cell, nn] for [i, j, cell, nn] in self.system.bonds]
            model_device.bonds = system_bonds + left_bonds + right_bonds
            model_device.initialize_hamiltonian(find_bonds=False)

        device_hamiltonian = sp.csc_matrix(model_device.hamiltonian[0].shape, dtype=np.complex_)
        for h in model_device.hamiltonian:
            device_hamiltonian += h

        self.bonds.append(model_device.bonds)

        left_lead_H  = self.__lead_hamiltonian(self.left_lead, -self.period[0])
        right_lead_H = self.__lead_hamiltonian(self.right_lead, self.period[1])

        return (device_hamiltonian, left_lead_H, right_lead_H)


    def __lead_hamiltonian(self, lead: np.ndarray, period: float) -> Tuple[sp.csc_matrix, sp.csc_matrix]:
        """
        Routine to attach together two lead unit cells to extract the cell Hamiltonian,
        and the coupling between cells.

        :param lead: Array with atomic positions and species of the lead.
        :param period: Norm of Bravais vector of the lead.
        :return: [H, V] Hamiltonian of unit cell and coupling between unit cells.2
        """

        displaced_lead = np.copy(lead)
        displaced_lead[:, :3] += np.array([period, 0, 0])
        if period > 0:
            left_lead_motif = np.concatenate((lead, displaced_lead))
        else:
            left_lead_motif = np.concatenate((displaced_lead, lead))
        model_lead = deepcopy(self.system)
        model_lead.motif = left_lead_motif
        model_lead.matrix_type = "sparse"
        if self.mode == "direct":
            max_nn = max(int(value) for value in self.system.configuration["SK amplitudes"].keys())
            model_lead.find_neighbours(mode="minimal", nn=max_nn)
            findBonds = False
        else:
            findBonds = True
        model_lead.initialize_hamiltonian(find_bonds=findBonds)
        lead_total_hamiltonian = sp.csc_matrix(model_lead.hamiltonian[0].shape, dtype=np.complex_)
        for h in model_lead.hamiltonian:
            lead_total_hamiltonian += h

        lead_dim = model_lead.basisdim // 2
        lead_coupling = lead_total_hamiltonian[:lead_dim, lead_dim:]
        lead_hamiltonian = lead_total_hamiltonian[:lead_dim, :lead_dim]

        self.bonds.append(model_lead.bonds)

        return (lead_hamiltonian, lead_coupling)


    def __lead_selfenergy(self, energy: float, lead_hamiltonian: sp.csr_matrix, lead_coupling: sp.csr_matrix, delta: float = 1E-7, 
                          threshold: float = 1E-10, mixing: float = 0.5, method: str = "LS"):
        """
        Routine to compute the selfenergy of one leaf at a given energy. 

        :param energy: Energy where we evaluate the selfenergy.
        :param lead_hamiltonian: Hamiltonian block corresponding to the unit cell of the lead.
        :param lead_coupling: Coupling between lead unit cells.
        :param delta: Imaginary part of lead. Usually infinitesimal, its sign defined retarded or advanced lead. Defaults to 1E-7.
        :param threshold: Threshold to stop self-consistency of selfenergy. Defaults to 1E-4.
        :param mixing: Mixing parameter to achieve convergence faster.
        :param method: Method to convergence the self-energies of the leads. Options are 'LS' (Lopez Sancho method)
            and 'LM' (Linear Mixing). Defaults to 'LS'.
        :return: Selfenergy matrix evaluated at energy+i*delta
        """

        notConverged = True
        identity = np.eye(lead_hamiltonian.shape[0])

        if method == "LS":
            s = lead_hamiltonian.todense()
            e = lead_hamiltonian.todense()
            alpha = lead_coupling.todense()
            beta = alpha.transpose().conjugate()
            
            while notConverged:
                g = np.linalg.inv((energy + 1j*delta)*identity - e)
                factor = alpha @ g @ beta
                s += factor
                e += factor + beta @ g @ alpha
                alpha = alpha @ g @ alpha
                beta = beta @ g @ beta
                norm = np.linalg.norm(alpha)
                if norm < threshold:
                    notConverged = False

            g = np.linalg.inv((energy + 1j*delta)*identity - s)
            selfenergy = lead_coupling @ g @ lead_coupling.transpose().conjugate()

        elif method == "LM":
            selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)

            old_selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)
            old2_selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)
            identity = sp.eye(lead_hamiltonian.shape[0], format="csc")
            while notConverged:
                old2_selfenergy = old_selfenergy
                old_selfenergy = selfenergy
                new_selfenergy = (lead_coupling @ 
                                sp.linalg.inv((energy + 1j*delta)*identity - lead_hamiltonian - selfenergy) @ 
                                lead_coupling.transpose().conjugate())
                selfenergy = old_selfenergy * (1 - mixing) + mixing * new_selfenergy 
                maxNorm = sp.linalg.norm(selfenergy - old_selfenergy)
                maxNorm2 = sp.linalg.norm(selfenergy - old2_selfenergy)
                if maxNorm < threshold and maxNorm2 < threshold:
                    notConverged = False
        
        else:
            raise ValueError("method must be either 'LS' or 'LM'.")

        return sp.csc_matrix(selfenergy)


    # def __lead_selfenergy(energy: float, lead_hamiltonian: sp.csr_matrix, lead_coupling: sp.csr_matrix, delta: float = 1E-7, threshold: float = 1E-3, mixing: float = 0.5):
    #     """
    #     Routine to compute the selfenergy of one leaf at a given energy. Implements a Pulay mixer to achieve self-consistency.

    #     :param energy: Energy where we evaluate the selfenergy.
    #     :param lead_hamiltonian: Hamiltonian block corresponding to the unit cell of the lead.
    #     :param lead_coupling: Coupling between lead unit cells.
    #     :param delta: Imaginary part of lead. Usually infinitesimal, its sign defined retarded or advanced lead. Defaults to 1E-7.
    #     :param threshold: Threshold to stop self-consistency of selfenergy. Defaults to 1E-4.
    #     :param mixing: Mixing parameter to achieve convergence faster.
    #     :return: Selfenergy matrix evaluated at energy+i*delta
    #     """

    #     selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)
    #     residuals, selfenergies = [], []

    #     notConverged = True
    #     old_selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)
    #     identity = sp.eye(lead_hamiltonian.shape[0], format="csc")
    #     n = 3
    #     first_iterations = 10
    #     counter = 0

    #     while notConverged:
            
    #         new_selfenergy = (lead_coupling @ 
    #                           sp.linalg.inv((energy + 1j*delta)*identity - lead_hamiltonian - selfenergy) @ 
    #                           lead_coupling.transpose().conjugate())
            
            
    #         # #  Pulay mixing
    #         # Store first n elements
    #         selfenergies.append(selfenergy)
    #         residuals.append(new_selfenergy - selfenergy)
    #         if counter < first_iterations:
    #             if counter >= n:
    #                 selfenergies.pop(0)
    #                 residuals.pop(0)
    #             selfenergy = mixing * selfenergy + (1 - mixing) * new_selfenergy
    #             norm = sp.linalg.norm(residuals[-1])
    #         # After storing first iterations, perform mixing
    #         else:
    #             selfenergies.pop(0)
    #             residuals.pop(0)
            
    #             residual_matrix = np.zeros([len(residuals) + 1, len(residuals) + 1], dtype=np.complex_)
    #             for i in range(len(residuals)):
    #                 ri = residuals[i].transpose().conjugate()
    #                 for j in range(len(residuals) - i):
    #                     rj = residuals[j]
    #                     if j == 0:
    #                         residual_matrix[i, i + j] = (ri @ rj).trace() / 2
    #                     else:
    #                         residual_matrix[i, i + j] = (ri @ rj).trace()
    #             residual_matrix += np.conjugate(np.transpose(residual_matrix))
    #             residual_matrix[:, -1] = 1
    #             residual_matrix[-1, :] = 1
    #             residual_matrix[-1, -1] = 0

    #             independent_vector = np.zeros(len(residuals) + 1)
    #             independent_vector[-1] = 1

    #             coefs = np.linalg.solve(residual_matrix, independent_vector)

    #             selfenergy = sp.csc_matrix(lead_hamiltonian.shape, dtype=np.complex_)
    #             for i, coef in enumerate(coefs[:-1]):
    #                 selfenergy += coef  * (selfenergies[i] + 0.5 * residuals[i])

    #             norm = np.sqrt(abs(coefs[-1]))
                    
    #         counter += 1
    #         if norm < threshold:
    #             notConverged = False
                
        
    #     return selfenergy


    def __device_green_function(self, energy: float, device: list, left_self_energy: sp.csc_matrix, right_self_energy: sp.csc_matrix, delta: float = 1E-7):
        """
        Routine to compute the Green's function of the device at energy + i*delta.

        :param energy: Energy where we evaluate the Green's function.
        :param device: List with the Fock matrices of the device.
        :param left_self_energy: Self-energy of left lead.
        :param right_self_energy: Self-energy of right lead.
        :param delta: Broadening. Defaults to 1E-7.
        :return: Green's function of the device evaluated at energy + i*delta.
        """

    
        device_dim = device.shape[0]
        
        extended_lead_selfenergy_L = sp.lil_matrix((device_dim, device_dim), dtype=np.complex_)
        extended_lead_selfenergy_L[:left_self_energy.shape[0], :left_self_energy.shape[0]] = left_self_energy

        extended_lead_selfenergy_R= sp.lil_matrix((device_dim, device_dim), dtype=np.complex_)
        extended_lead_selfenergy_R[(device_dim - right_self_energy.shape[0]):, (device_dim - right_self_energy.shape[0]):] = right_self_energy

        extended_lead_selfenergy_L = extended_lead_selfenergy_L.tocsc()
        extended_lead_selfenergy_R = extended_lead_selfenergy_R.tocsc()

        green_device = sp.linalg.inv((energy + 1j*delta)*sp.eye(device_dim, format="csc") - device - extended_lead_selfenergy_L - extended_lead_selfenergy_R)

        return green_device


    def visualize_device(self, ax: Axes = None, pbc: bool = False, ) -> None:
        """
        Routine to visualize the device. Intended to be used only with two dimensional (planar) systems.

        :param ax: Axis object to plot the device on.
        :param pbc: Boolean to toggle 3d plot of the device in case the system and the leads have PBC.
        """

        if ax is None:
            if pbc and self.system.boundary == "PBC":
                ax = plt.figure().add_subplot(projection='3d')
                # fig, ax = plt.subplots(1, 1, subplot_kw={"projection": "3d"})
            elif pbc and self.system.boundary == "OBC":
                    print("Warning: plot set to periodic but system is finite. Plot will be 2d.")
            else:
                fig, ax = plt.subplots(1, 1)


        # First plot bonds
        displaced_left_motif = np.copy(self.left_lead)
        displaced_left_motif[:, :3] -= [self.period[0], 0, 0]

        displaced_right_motif = np.copy(self.right_lead)
        displaced_right_motif[:, :3] += [self.period[1], 0, 0]

        left_lead = np.concatenate((displaced_left_motif, self.left_lead))
        right_lead = np.concatenate((self.right_lead, displaced_right_motif))

        system_motif = self.system.motif

        # Map into cilinder. z component is disregarded.
        if pbc:
            L = self.system.bravais_lattice[0][1]
            r = L/(2*np.pi)
            for n, atom in enumerate(left_lead):
                y = atom[1]
                left_lead[n, 1] = r*np.cos(2*np.pi*y/L)
                left_lead[n, 2] = r*np.sin(2*np.pi*y/L)
            for n, atom in enumerate(right_lead):
                y = atom[1]
                right_lead[n, 1] = r*np.cos(2*np.pi*y/L)
                right_lead[n, 2] = r*np.sin(2*np.pi*y/L)
        
        if not pbc:
            # Then plot atoms
            for atom in system_motif:
                ax.scatter(atom[0], atom[1], c="mediumseagreen", edgecolors="black", linewidths=1)
            
            for atom in self.left_lead:
                ax.scatter(atom[0], atom[1], c="orange", edgecolors="black", linewidths=1)
                ax.scatter(atom[0] - self.period[0], atom[1], c="moccasin", edgecolors="black", linewidths=1)

            for atom in self.right_lead:
                ax.scatter(atom[0], atom[1], c="orange", edgecolors="black", linewidths=1)
                ax.scatter(atom[0] + self.period[0], atom[1], c="moccasin", edgecolors="black", linewidths=1)
            
            total_motif = np.concatenate((self.left_lead, self.system.motif, self.right_lead), axis=0)
            for bond in self.bonds[0]:
                rx = [total_motif[bond[0], 0], total_motif[bond[1], 0]]
                ry = [total_motif[bond[0], 1], total_motif[bond[1], 1]]
                ax.plot(rx, ry, "k-", zorder=-1)
            
            for bond in self.bonds[1]:
                rx = [left_lead[bond[0], 0], left_lead[bond[1], 0]]
                ry = [left_lead[bond[0], 1], left_lead[bond[1], 1]]
                ax.plot(rx, ry, "k-", zorder=-1)
            
            for bond in self.bonds[2]:
                rx = [right_lead[bond[0], 0], right_lead[bond[1], 0]]
                ry = [right_lead[bond[0], 1], right_lead[bond[1], 1]]
                ax.plot(rx, ry, "k-", zorder=-1)

            ax.axis('equal')

        else:
            for atom in system_motif:
                ax.scatter(atom[0], r*np.cos(2*np.pi*atom[1]/L), r*np.sin(2*np.pi*atom[1]/L), c="mediumseagreen", edgecolors="black", linewidths=1)
            
            for atom in self.left_lead:
                ax.scatter(atom[0], r*np.cos(2*np.pi*atom[1]/L), r*np.sin(2*np.pi*atom[1]/L), c="orange", edgecolors="black", linewidths=1)
                ax.scatter(atom[0] - self.period[0], r*np.cos(2*np.pi*atom[1]/L), r*np.sin(2*np.pi*atom[1]/L), c="moccasin", edgecolors="black", linewidths=1)

            for atom in self.right_lead:
                ax.scatter(atom[0], r*np.cos(2*np.pi*atom[1]/L), r*np.sin(2*np.pi*atom[1]/L), c="orange", edgecolors="black", linewidths=1)
                ax.scatter(atom[0] + self.period[0], r*np.cos(2*np.pi*atom[1]/L), r*np.sin(2*np.pi*atom[1]/L), c="moccasin", edgecolors="black", linewidths=1)
            
            total_motif = np.concatenate((self.left_lead, self.system.motif, self.right_lead), axis=0)
            for bond in self.bonds[0]:
                rx = [total_motif[bond[1], 0], total_motif[bond[0], 0]]
                ry_org = [total_motif[bond[1], 1], total_motif[bond[0], 1]]
                ry = [r*np.cos(2*np.pi*value/L) for value in ry_org]
                rz = [r*np.sin(2*np.pi*value/L) for value in ry_org]
                ax.plot(rx, ry, rz, "k-", zorder=-1)
            
            for bond in self.bonds[1]:
                rx = [left_lead[bond[1], 0], left_lead[bond[0], 0]]
                ry = [left_lead[bond[1], 1], left_lead[bond[0], 1]]
                rz = [left_lead[bond[1], 2], left_lead[bond[0], 2]]
                ax.plot(rx, ry, rz, "k-", zorder=-1)
            
            for bond in self.bonds[2]:
                rx = [right_lead[bond[1], 0], right_lead[bond[0], 0]]
                ry = [right_lead[bond[1], 1], right_lead[bond[0], 1]]
                rz = [right_lead[bond[1], 2], right_lead[bond[0], 2]]
                ax.plot(rx, ry, rz, "k-")

            ax.view_init(elev=0)
            try:
                ax.set_aspect('equal')
            except NotImplementedError as e:
                print("Plotting the device in 3d requires matpltolib 3.7.0 or higher.")
                raise(e)
            ax.set_axis_off()



def ldos(result: Spectrum, atom_index: int, energy: float, delta: float = 1E-4):
    """
    Routine to compute the local density of states at a given energy.

    :param result: Spectrum object with the results from the diagonalization.
    :param atom_index: Index of the atom where we want to compute the LDOS.
    :param energy: Value of energy where we evaluate the LDOS.
    :param delta: Value of energy broadening. Defaults to 1E-4.
    :return: Value of LDOS.
    """

    pass


def integrated_ldos(result: Spectrum, atom_index: int, minE: float, maxE: float):
    """
    Routine to compute the local density of states integrated on a specified energy window.

    :param result: Spectrum object with the result of the diagonalization of the system.
    :param atom_index: Index of the atom where we evaluate the integrated LDOS.
    :param minE: Minimum value of energy window.
    :param maxE: Maximum value of energy window.
    :return: Integrated LDOS evaluated on atom_index.
    """

    index_to_matrix_array = [0]
    for index in range(result.system.natoms - 1):
        species = int(result.system.motif[index, 3])
        index_to_matrix_array.append(index_to_matrix_array[-1] + result.system.norbitals[species])

    init_index = int(index_to_matrix_array[atom_index])
    species_index = int(result.system.motif[atom_index, 3])
    final_index = init_index + int(result.system.norbitals[species_index])

    ildos = 0
    for j in range(result.eigen_energy.shape[1]):
        for i in range(result.eigen_energy.shape[0]):
            eigval = result.eigen_energy[i, j]
            if eigval > maxE or eigval < minE:
                continue
            eigvec = result.eigen_states[j, :, i]
            eigvec = eigvec[init_index:final_index]
            ildos += np.vdot(eigvec, eigvec)
    
    return ildos





from tightbinder.observables import dos, dos_kpm, TransportDevice
from tightbinder.fileparse import parse_config_file
from tightbinder.models import SlaterKoster
from tightbinder.utils import hash_numpy_array
import numpy as np
import math

"""
Module with tests to verify the functionality of the observables module.
"""


def test_dos():
    """
    Tests calculation of the density of states.
    """
    
    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)
    model = SlaterKoster(config)

    nk = 20
    kpoints = model.brillouin_zone_mesh([nk, nk])

    model.initialize_hamiltonian()
    results = model.solve(kpoints)

    density, energies = dos(results, delta=0.05, npoints=200)
    
    density_hash = hash_numpy_array(np.array(density))
    
    assert math.isclose(density_hash, 5.678625492981987)
            

def test_dos_kpm():
    """
    Tests calculation of the density of states using the KPM.
    """
    
    np.random.seed(1)

    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)

    ncells = 10
    model = SlaterKoster(config).supercell(n1=ncells, n2=ncells)

    model.matrix_type = "sparse"
    model.initialize_hamiltonian()

    density, energies = dos_kpm(model, nmoments=150, npoints=400, r=30)
    
    density_hash = hash_numpy_array(np.array(density))
    
    assert math.isclose(density_hash, 0.7510637669286024)
            

def test_transport_device():
    """
    Tests initialization of a transport device.
    """
    
    length, width = 10, 5

    file = open("examples/chain.txt", "r")
    config = parse_config_file(file)

    model = SlaterKoster(config)

    model.bravais_lattice = np.concatenate((model.bravais_lattice, np.array([[0., 1, 0]])))
    model = model.reduce(n2=width)

    left_lead = np.copy(model.motif)
    left_lead[:, :3] -= model.bravais_lattice[0]

    right_lead = np.copy(model.motif)
    right_lead[: , :3] += length * model.bravais_lattice[0]

    period = model.bravais_lattice[0, 0]

    model = model.reduce(n1=length)
    
    device = TransportDevice(model, left_lead, right_lead, period, "default")
        
    
def test_transmission():
    """
    Tests calculation of the transmission.
    """
    
    length, width = 7, 4

    file = open("examples/chain.txt", "r")
    config = parse_config_file(file)

    model = SlaterKoster(config)

    model.bravais_lattice = np.concatenate((model.bravais_lattice, np.array([[0., 1, 0]])))
    model = model.reduce(n2=width)

    left_lead = np.copy(model.motif)
    left_lead[:, :3] -= model.bravais_lattice[0]

    right_lead = np.copy(model.motif)
    right_lead[: , :3] += length * model.bravais_lattice[0]

    period = model.bravais_lattice[0, 0]

    model = model.reduce(n1=length)
    
    device = TransportDevice(model, left_lead, right_lead, period, "default")
    trans, energy = device.transmission(-5, 5, 50)
    
    transmission_hash = hash_numpy_array(np.array(trans))
    
    assert math.isclose(transmission_hash, 243.99797881099673)
        

def test_conductance():
    """
    Tests calculation of the conductance.
    """
    
    length, width = 7, 4

    file = open("examples/chain.txt", "r")
    config = parse_config_file(file)

    model = SlaterKoster(config)

    model.bravais_lattice = np.concatenate((model.bravais_lattice, np.array([[0., 1, 0]])))
    model = model.reduce(n2=width)

    left_lead = np.copy(model.motif)
    left_lead[:, :3] -= model.bravais_lattice[0]

    right_lead = np.copy(model.motif)
    right_lead[: , :3] += length * model.bravais_lattice[0]

    period = model.bravais_lattice[0, 0]

    model = model.reduce(n1=length)
    
    device = TransportDevice(model, left_lead, right_lead, period, "default")
    conductance = device.conductance()
                
    assert np.allclose(conductance, 6.064302600409135)




""" 
Routines to parse the tight-binding parameters written in the input file.
"""

from io import TextIOWrapper
import re
from tightbinder.utils import pretty_print_dictionary
from typing import List

def parse_raw_arguments(file: TextIOWrapper) -> dict:
    """ 
    Routine to parse the arguments raw (without any treatment) from the input file, following
    the rules defined for configuration files.

    :param file: Pointer to configuration file, obtained from call to open().
    :return: Dictionary with the content corresponding to each flag of the config. file. 
    """

    arguments = {}
    value = []

    lines = file.readlines()
    for line in lines:
        # Identifier
        if line[0] == '#':
            try:
                arguments.update({new_argument: value})
                value = []
            except UnboundLocalError:
                pass
            new_argument = line[1:].strip()
        # Blank line
        elif line[0] in ['\n', '\r\n', '!', ' ']:
            continue
        # Argument values
        else:
            value.append(line.strip())
    arguments.update({new_argument: value})

    return arguments


def check_arguments(arguments: dict, required_arguments: List[str]) -> None:
    """ 
    Routine to check whether all required fields for model are present in the input file.

    :param arguments: Dictionary whose keys are the arguments to compare.
    :param required_arguments: List with expected arguments.
    :raises KeyError: Raise exception if not all required arguments are present in the arguments. 
    """

    for arg in required_arguments:
        if arg not in arguments.keys():
            raise KeyError(f"{arg} not present in input file")


def shape_arguments(arguments: dict) -> dict:
    """ Routine to rewrite correctly the arguments parsed (raw) from the input file.

    :param arguments: Dictionary with the raw contents of the config. file.
    :return: Dictionary with the contents of the config. file in numerical format. 
    :raises IndexError: Raised if some arguments are missing
    :raises ValueError: Raised if some arguments have incorrect values (e.g. string instead of numbers).
    :raises SyntaxError: Raised for the SK amplitudes if not using correctly the brackets.
    :raises NotImplementedError: Raised if there is an unexpected argument present. 
    """

    for arg in arguments:
        if arg == 'System name':
            try:
                arguments[arg] = arguments[arg][0]
            except IndexError as e:
                print(f'{type(e).__name__}: No system name given')
                raise

        elif arg in ['Dimensionality', 'Species']:
            try:
                arguments[arg] = int(arguments[arg][0])
            except IndexError as e:
                print(f'{type(e).__name__}: No {arg} given')
                raise
            except ValueError as e:
                print(f'{type(e).__name__}: {arg} has to be an integer')
                raise

        elif arg == 'Bravais lattice':
            aux_array = []
            for line in arguments[arg]:
                try:
                    aux_array.append([float(num) for num in re.split(' |, |,', line)])
                except IndexError as e:
                    print(f'{type(e).__name__}: No {arg} vectors given given')
                    raise
                except ValueError as e:
                    print(f'{type(e).__name__}: {arg} vectors have to be numbers')
                    raise
            arguments[arg] = aux_array

        elif arg == 'Motif':
            aux_array = []
            for n, line in enumerate(arguments[arg]):
                aux_array.append([float(num) for num in re.split(' |, |,|; |;', line)])
                if arguments['Species'] == 1:
                    try:
                        aux_array[n][3] = 0  # Default value to 1
                    except IndexError:
                        aux_array[n].append(0)
            arguments[arg] = aux_array

        elif arg == 'Orbitals':
            possible_orbitals = ['s', 'px', 'py', 'pz', 'dxy', 'dyz', 'dzx', 'dx2-y2', 'd3z2-r2']
            aux_array = []
            for line in arguments[arg]:
                try:
                    orbitals = re.split(' |, |,', line)
                except IndexError as e:
                    print(f'{type(e).__name__}: No orbitals included')
                    raise
                else:
                    # Check that all are correctly written
                    for orbital in orbitals:
                        if orbital not in possible_orbitals:
                            print(orbital)
                            print('Error: Incorrect orbital specified')
                            raise ValueError("Specified orbitals are not correct")
                    aux_array.append(orbitals)
            arguments[arg] = aux_array

        elif arg == 'Onsite energy':
            aux_array = []
            for line in arguments[arg]:
                try:
                    aux_array.append([float(num) for num in re.split(' |, |,', line)])
                except IndexError as e:
                    print(f'{type(e).__name__}: No onsite energies included')
                    raise
                except ValueError as e:
                    print(f'{type(e).__name__}: Onsite energies must be numbers')
                    raise
            arguments[arg] = aux_array

        elif arg == 'SK amplitudes':
            dictionary = {}
            for n, line in enumerate(arguments[arg]):
                if not line:
                    raise SyntaxError('No Slater-Koster amplitudes given')
                else:
                    # First try to parse square brakets
                    if ((line.find('[') != -1 and line.find(']') == -1) 
                        or 
                        (line.find('[') == -1 and line.find(']') != -1)):
                        raise SyntaxError('SK amplitudes: Square brackets must be closed')
                    line = list(filter(None, re.split(r'\[|\]', line)))
                    if len(line) == 1 and len(arguments[arg]) == 1:
                        species_neigh = ['0', '0', '1']
                    elif len(line) == 1 and len(arguments[arg]) != 1:
                        raise SyntaxError(f'{arg}: Only one line can be present if not specifying species or neighbours')
                    elif len(line) == 2:
                        species_neigh = re.split(' |, |,|; |;', line[0])
                        if len(species_neigh) == 1:
                            species_neigh = ['0', '0'] + species_neigh
                        elif len(species_neigh) == 2:
                            species_neigh = species_neigh + ['1']
                        elif len(species_neigh) > 3:
                            raise SyntaxError(f'{arg}: More values than expected in braket')     
                    else:
                        raise SyntaxError(f'{arg}: Too many square brackets found')

                    neigh = species_neigh[2]                
                    if neigh not in dictionary:
                        dictionary[species_neigh[2]] = {}

                    amplitudes = list(filter(None, re.split(' |, |,|; |;', line[-1].strip())))
                    try:
                        amplitudes = [float(num) for num in amplitudes]
                        species = species_neigh[0] + species_neigh[1]
                        if species in dictionary[neigh]:
                            print(f'\033[93m Warning: Overwriting SK amplitudes for species pair ' + species + '\033[0m')
                        dictionary[neigh][species] = amplitudes
                    except ValueError as e:
                        raise Exception(f'Slater-Koster amplitudes must be numbers ({amplitudes})') from e

            arguments[arg] = dictionary

        elif arg == 'Spin':
            if arguments[arg][0] == "True":
                arguments[arg] = True
            elif arguments[arg][0] == "False":
                arguments[arg] = False
            else:
                print('Error: Spin parameter must be True or False (or 1 or 0 respectively)')
                raise

        elif arg == 'Spin-orbit coupling':
            aux_array = []
            for line in arguments[arg]:
                try:
                    aux_array.append(float(line))
                except IndexError as e:
                    print(f'Warning: No spin-orbit coupling given, defaulting to 0...')
                    arguments[arg] = 0.0
                except ValueError as e:
                    print(f'{type(e).__name__}: Spin-orbit coupling must be a number')
                    raise
            arguments[arg] = aux_array

        elif arg == 'Filling':
            aux_array = []
            for line in arguments[arg]:
                try:
                    aux_array.append(float(line))
                except ValueError as e:
                    print(f'{type(e).__name__}: Filling must be a number')
                    raise
            arguments[arg] = aux_array

        elif arg == 'Mesh':
            try:
                arguments[arg] = [int(num) for num in re.split(' |, |,', arguments[arg][0])]
            except IndexError as e:
                print(f'{type(e).__name__}: No mesh given')
                raise
            except ValueError as e:
                print(f'{type(e).__name__}: Mesh must be integer numbers')
                raise

        elif arg in ['Radius', 'Mixing']:
            try:
                arguments[arg] = float(arguments[arg][0])
            except IndexError as e:
                print(f'{type(e).__name__}: No {arg} given')
                raise
            except ValueError as e:
                print(f'{type(e).__name__}: {arg} has to be a float')
                raise

        elif arg == 'High symmetry points':
            try:
                arguments[arg] = [str(label) for label in re.split(' |, |,', arguments[arg][0])]
            except IndexError as e:
                print(f'{type(e).__name__}: No high symmetry points given')
                raise

        else:
            raise NotImplementedError(f'{arg} is not a valid parameter')

    return arguments


def check_coherence(arguments: dict) -> None:
    """ 
    Routine to check that the present arguments are coherent among them.

    :param arguments: Dictionary with the config. file content already processed with shape_arguments(). 
    """

    # --------------- Model ---------------
    # Check dimensions
    if arguments['Dimensionality'] > 3 or arguments['Dimensionality'] < 0:
        raise ValueError('Error: Invalid dimension!')

    # Check species
    if arguments['Species'] < 0:
        raise ValueError('Error: Species has to be a positive number (1 or 2)')

    # Check vector basis
    if arguments['Dimensionality'] != len(arguments['Bravais lattice']):
        raise AssertionError('Error: Dimension and number of basis vectors do not match')

    # Check length of vector basis
    for vector in arguments['Bravais lattice']:
        if len(vector) != 3:
            raise ValueError('Error: Bravais vectors must have three components')

    # Check that motif has elements specified if num. of species = 2
    if arguments['Species'] == 2:
        for atom in arguments['Motif']:
            try:
                atom_element = atom[3]  # Try to access
            except IndexError:
                print('Error: Atomic species not specified in motif')
                raise

            if atom_element > arguments['Species']:
                raise AssertionError('Incorrect species labeling in motif')

    # Check onsite energies are present for all species
    if len(arguments['Onsite energy']) > arguments['Species']:
        raise AssertionError('Too many onsite energies for given species')
    elif len(arguments['Onsite energy']) < arguments['Species']:
        raise AssertionError('Missing onsite energies for both atomic species')

    # Check orbitals present for all species
    if (len(arguments['Orbitals']) != arguments['Species']):
        raise AssertionError("Orbitals must be specified for each species")

    # Check number of SK coefficients
    for neighbour in arguments["SK amplitudes"].keys():
        nspecies_pairs = len(arguments['SK amplitudes'][neighbour].keys())
        if nspecies_pairs > arguments['Species']*(arguments['Species'] + 1)/2:
            raise AssertionError('Too many rows of SK coefficients')

    # Check number of SOC strenghts
    if arguments['Spin'] and len(arguments['Spin-orbit coupling']) != arguments["Species"]:
        raise AssertionError("Values for SOC strength must match number of species")

    # Check if SOC are non-zero
    for soc in arguments['Spin-orbit coupling']:
        if soc != 0 and not arguments['Spin']:
            print('Warning: Spin-orbit coupling is non-zero but spin was set to False. ')
            arguments['Spin'] = True
    
    # Check if mixing has a valid value if it is present
    if 'Mixing' in arguments:
        if arguments["Mixing"] < 0 or arguments["Mixing"] > 1:
            raise AssertionError("Mixing must be a value between 0 and 1")

        # TODO: Check if this is necessary
        #if not all([len(orbitals)==len(arguments["Orbitals"][0]) for orbitals in arguments["Orbitals"]]):
        #    raise AssertionError("Mixing can only be used with isoelectronic species")
    
    if 'Radius' in arguments:
        if arguments['Radius'] < 0:
            raise AssertionError("Radius must be a positive number")

    if 'Filling' in arguments:
        if len(arguments['Filling']) != arguments['Species']:
            raise AssertionError("Must provide number of electrons of each chemical species.")

        total_electrons = 0.0
        for atom in arguments['Motif']:
            species = int(atom[3])
            total_electrons += arguments['Filling'][species]
        if not total_electrons.is_integer():
            raise AssertionError('Total number of electrons of the system must be a positive integer.')

    # ------------ Orbital consistency ------------
    # Check onsite energy per orbital per species
    for n, onsite_energies in enumerate(arguments["Onsite energy"]):
        if len(onsite_energies) != len(arguments["Orbitals"][n]):
            raise AssertionError("Error: Each orbital requires one onsite energy value")

    # Check SK amplitudes match present orbitals
    for neighbour in arguments["SK amplitudes"].keys():
        for species, coefs in arguments["SK amplitudes"][neighbour].items():
            needed_SK_coefs = 0

            first_orbitals = arguments["Orbitals"][int(species[0])]
            second_orbitals = arguments["Orbitals"][int(species[1])]

            first_orbital_list = []
            for orbital in first_orbitals:
                if orbital[0] not in first_orbital_list:
                    first_orbital_list.append(orbital[0])
            
            second_orbital_list = []
            for orbital in second_orbitals:
                if orbital[0] not in second_orbital_list:
                    second_orbital_list.append(orbital[0])
            
            amplitudes_per_orbitals = {"ss": 1, "sp": 1, "sd": 1, "pp": 2, "pd": 2, "dd":3}
            for key in amplitudes_per_orbitals.keys():
                if((key[0] in first_orbital_list and key[1] in second_orbital_list)
                    or 
                (key[1] in first_orbital_list and key[0] in second_orbital_list)
                ):
                    needed_SK_coefs += amplitudes_per_orbitals[key]

            if needed_SK_coefs != len(coefs):
                raise AssertionError("Wrong number of SK amplitudes for given orbitals")


            # Check whether all necessary SK coefs are present for all orbitals
            if len(coefs) != needed_SK_coefs:
                raise AssertionError('Mismatch between orbitals and required SK amplitudes')

    # ---------------- Simulation ----------------
    # Check mesh matches dimension
    if len(arguments['Mesh']) != arguments['Dimensionality']:
        raise AssertionError('Mesh dimension does not match system dimension')

    if 'Radius' in arguments and len(arguments['SK amplitudes'].keys()) > 1:
        raise AssertionError('Must not specify neighbours in radius mode')



def sk_coefficients_standard_form(coefficients: list, orbitals: list) -> list:
    """ 
    Routine to transform the SK coefficients from the parsed form to the
    standard form.

    :param coefficients: List with the parsed coefficients
    :param orbitals: List with orbitals corresponding to the given SK amplitudes
    :return: List with SK amplitudes in standard form. 
    """

    pass


def transform_sk_coefficients(configuration: dict) -> None:
    """ 
    Routine to transform SK coefficients into standardized form
    for later manipulation in hamiltonian.

    :param configuration: Dictionary with the contents of the config. file already shaped. 
    """

    # First store untransformed amplitudes.
    configuration['SK'] = configuration["SK amplitudes"]

    dict = {}
    for neighbour in configuration["SK amplitudes"].keys():
        dict[neighbour] = {}
        for species, coefs in configuration["SK amplitudes"][neighbour].items():
            amplitudes = [0]*10  # Number of possible SK amplitudes
            first_orbitals = configuration["Orbitals"][int(species[0])]
            second_orbitals = configuration["Orbitals"][int(species[1])]

            first_orbital_list = []
            for orbital in first_orbitals:
                if orbital[0] not in first_orbital_list:
                    first_orbital_list.append(orbital[0])
            
            second_orbital_list = []
            for orbital in second_orbitals:
                if orbital[0] not in second_orbital_list:
                    second_orbital_list.append(orbital[0])
            
            if "d" in second_orbital_list and "d" not in first_orbital_list:
                aux_array = second_orbital_list
                second_orbital_list = first_orbital_list
                first_orbital_list = aux_array
            if "d" not in first_orbital_list and "d" not in second_orbital_list:
                if "p" in second_orbital_list and "p" not in first_orbital_list:
                    aux_array = second_orbital_list
                    second_orbital_list = first_orbital_list
                    first_orbital_list = aux_array

            it = 0
            if "s" in first_orbital_list:
                if "s" in second_orbital_list:
                    amplitudes[0] = coefs[it]
                    it += 1
            if "p" in first_orbital_list:
                if "s" in second_orbital_list:
                    amplitudes[1] = coefs[it]
                    it += 1
                if "p" in second_orbital_list:
                    amplitudes[2:4] = coefs[it:it+2]
                    it += 2
            if "d" in first_orbital_list:
                if "s" in second_orbital_list:
                    amplitudes[4] = coefs[it]
                    it += 1
                if "p" in second_orbital_list:
                    amplitudes[5:7] = coefs[it:it+2]
                    it += 2
                if "d" in second_orbital_list:
                    amplitudes[7:] = coefs[it:]

            dict[neighbour][species] = amplitudes
        
    configuration['SK amplitudes'] = dict



def mix_parameters(configuration: dict) -> None:
    """ 
    Method to mix the given parameters in case that some are missing in presence of multiple species.

    :param configuration: Dictionary with the contents of the config. file already shaped. 
    """

    if 'Mixing' not in configuration:
        mixing = 0.5
    else:
        mixing = configuration['Mixing']

    for i in range(configuration['Species']):
        for j in range(i + 1, configuration['Species']):
            species   = str(i) + str(j)
            species_i = str(i) + str(i)
            species_j = str(j) + str(j)
            for neighbour in configuration['SK amplitudes'].keys():
                SK_dictionary = configuration['SK amplitudes'][neighbour]
                if species not in SK_dictionary and species_i in SK_dictionary and species_j in SK_dictionary:
                    SK_dictionary[species] = [SK_dictionary[species_i][n]*mixing + SK_dictionary[species_j][n]*(1 - mixing) for n in range(len(SK_dictionary[species_i]))]


def parse_config_file(file: TextIOWrapper) -> dict:
    """ 
    Routine to obtain all the information from the configuration file, already shaped and verified.

    :param file: Python pointer to config. file as returned by open().
    :return: Dictionary with the contents of the config. file in standarized form, ready to be used by
        the class SlaterKoster. 
    """

    print("Parsing configuration file... ", end='')
    configuration = shape_arguments(parse_raw_arguments(file))
    required_arguments = ['System name', 'Dimensionality', 'Bravais lattice', 'Species',
                          'Motif', 'Orbitals', 'Onsite energy', 'SK amplitudes']
    check_arguments(configuration, required_arguments)
    check_coherence(configuration)

    mix_parameters(configuration)
    transform_sk_coefficients(configuration)
    
    print("Done\n")

    return configuration



""" 
Model declarations, from the Slater-Koster tight-binding model
to toy models such as the BHZ model or Wilson fermions.
"""

from __future__ import annotations
from typing import List, Tuple, Union

from .system import System, FrozenClass
from .crystal import Crystal
import numpy as np
import scipy.sparse as sp
import sys
import math
import cmath
from .utils import overrides

# Module-level variables
sigma_x = np.array([[0, 1],
                    [1, 0]], dtype=np.complex_)
sigma_y = np.array([[0, -1j],
                    [1j, 0]], dtype=np.complex_)
sigma_z = np.array([[1, 0],
                    [0, -1]], dtype=np.complex_)
"""
Definition of Pauli matrix x, y, z. Used by several predefined models.
"""

# ------------------ Constants and auxiliary methods ------------------
PI = 3.14159265359

# ------------------------------ Models ------------------------------


class SlaterKoster(System):
    """ 
    Implementation of Slater-Koster tight-binding model.
    
    :param configuration: Dictionary generated by fileparse.
    :param mode: Used to specify how to find the bonds. Either 'minimal' or
        'radius'. Minimal mode looks for first, second, ..., neighbours depending on
        the maximum distance specified on the configuration file. Radius mode defines
        bonds for all atoms within the specified radius. Defaults to 'minimal'.
    :param r: Value of the radius. Relevant only in 'radius' mode. Defaults to None.
    :param boundary: Boundary conditions, either 'PBC' or 'OBC'. Defaults to 'PBC'.
    """

    def __init__(self, configuration: dict = None, mode: str = 'minimal', r: float = None, boundary: str = "PBC") -> None:
        if configuration is not None:
            super().__init__(system_name=configuration["System name"],
                             bravais_lattice=configuration["Bravais lattice"],
                             motif=configuration["Motif"])
        else:
            super().__init__()

        self.configuration = configuration
        self.species = configuration['Species']
        self.boundary = boundary
        self._mode = mode
        self._r = r

        self.hamiltonian = None
        self.neighbours = None
        self.spin_orbit_hamiltonian = None        
        self.__zeeman = None
        self.__ordering = None
        self._unit_cell_list = None

        if mode not in ['minimal', 'radius']:
            print('Error: Incorrect mode')
            sys.exit(1)
        
        self.neighbours = len(configuration['SK amplitudes'].keys())

    @property
    def ordering(self):
        return self.__ordering

    @ordering.setter
    def ordering(self, ordering: str):
        if ordering != "atomic" and ordering != "spin":
            raise ValueError("ordering must be either atomic or spin")
        self.__ordering = ordering

    @property
    def nspecies(self):
        return self.configuration["Species"]

    @property
    def filling(self) -> int:
        if 'Filling' in self.configuration:
            nelectrons = 0
            for atom in self.motif:
                species = int(atom[3])
                nelectrons += self.configuration['Filling'][species]

            return int(nelectrons)


    # --------------- Methods ---------------
    def _hopping_amplitude(self, position_diff: np.ndarray, orbitals: list) -> float:
        """ 
        Routine to compute the hopping amplitude from one atom to another depending on the
        participating orbitals, following the Slater-Koster approximation. 

        :param position_diff: Array with the vector connecting the initial atom and the final atom.
        :param orbitals: List with [initial_orbital, initial_species, final_orbital, final_species, neighbour_index].
            neighbour_index denotes if they are first nn, second, etc.
        :return: Hopping amplitude.
        """

        initial_orbital = orbitals[0]
        initial_species = int(orbitals[1])
        final_orbital   = orbitals[2]
        final_species   = int(orbitals[3])
        neigh           = orbitals[4]

        initial_orbital_type = initial_orbital[0]
        final_orbital_type = final_orbital[0]

        amplitudes = self.configuration['SK amplitudes'][neigh]

        possible_orbitals = {'s': 0, 'p': 1, 'd': 2}
        if possible_orbitals[initial_orbital_type] > possible_orbitals[final_orbital_type]:
            position_diff = np.array(position_diff) * (-1)

            initial_orbital, final_orbital = final_orbital, initial_orbital
            initial_species, final_species = final_species, initial_species

            initial_orbital_type = initial_orbital[0]
            final_orbital_type = final_orbital[0]
            
        d_orbitals = {'dxy': 0, 'dyz': 0, 'dzx': 0, 'dx2-y2': 1, 'd3z2-r2':2}
        if (initial_orbital_type == "d" and final_orbital_type == "d"
                and
            d_orbitals[initial_orbital] > d_orbitals[final_orbital]):
            position_diff = np.array(position_diff) * (-1)
           
            initial_orbital, final_orbital = final_orbital, initial_orbital
            initial_species, final_species = final_species, initial_species

            initial_orbital_type = initial_orbital[0]
            final_orbital_type = final_orbital[0]
        
        species_pair = str(initial_species) + str(final_species)
        if species_pair not in amplitudes.keys():
            species_pair = str(final_species) + str(initial_species)

        effective_amplitudes = amplitudes[species_pair]
        direction_cosines = position_diff / np.linalg.norm(position_diff)
        direction_cosines = {'x': direction_cosines[0], 'y': direction_cosines[1], 'z': direction_cosines[2]}
        [l, m, n] = direction_cosines.values()
        (Vsss, Vsps, Vpps, Vppp, Vsds, Vpds, Vpdp, Vdds, Vddp, Vddd) = effective_amplitudes
        special_orbital = True if final_orbital == 'dx2-y2' or final_orbital == 'd3z2-r2' else False

        # Start checking the different possibilities
        if initial_orbital_type == "s":
            if final_orbital == "s":
                hopping = Vsss
            elif final_orbital_type == "p":
                coordinate_final = final_orbital[1]
                hopping = direction_cosines[coordinate_final] * Vsps
            elif final_orbital_type == "d" and not special_orbital:
                first_coordinate = final_orbital[1]
                second_coordinate = final_orbital[2]
                hopping = math.sqrt(3) * direction_cosines[first_coordinate] * direction_cosines[second_coordinate] * Vsds
            elif final_orbital == "dx2-y2":
                hopping = math.sqrt(3)/2.*(l*l - m*m)*Vsds
            elif final_orbital == "d3z2 - r2":
                hopping = (n*n - (l*l + m*m)/2.)*Vsds

        elif initial_orbital_type == "p":
            coordinate_initial = initial_orbital[1]
            if final_orbital_type == "p":
                coordinate_final = final_orbital[1]
                hopping = direction_cosines[coordinate_initial]*direction_cosines[coordinate_final]*(Vpps - Vppp)
                if coordinate_final == coordinate_initial:
                    hopping += Vppp
            elif final_orbital_type == "d" and not special_orbital:
                first_coordinate = final_orbital[1]
                second_coordinate = final_orbital[2]
                hopping = (direction_cosines[coordinate_initial]*direction_cosines[first_coordinate]
                           * direction_cosines[second_coordinate]*(math.sqrt(3)*Vpds - 2*Vpdp))
                if coordinate_initial == first_coordinate:
                    hopping += direction_cosines[second_coordinate]*Vpdp
                elif coordinate_initial == second_coordinate:
                    hopping += direction_cosines[first_coordinate]*Vpdp
            elif final_orbital == "dx2-y2":
                hopping = direction_cosines[coordinate_initial]*(l*l - m*m)*(math.sqrt(3)/2.*Vpds - Vpdp)
                if coordinate_initial == 'x':
                    hopping += l*Vpdp
                elif coordinate_initial == 'y':
                    hopping += -m*Vpdp
            elif final_orbital == "d3z2-r2":
                hopping = direction_cosines[coordinate_initial]*(n*n - (l*l + m*m)/2.)*Vpds
                if coordinate_initial == 'z':
                    hopping += math.sqrt(3)*n*(l*l + m*m)*Vpdp
                else:
                    hopping -= math.sqrt(3)*direction_cosines[coordinate_initial]*n*n*Vpdp

        elif initial_orbital_type == "d" and initial_orbital not in ["dx2-y2", "d3z2-r2"]:
            coordinate_initial_first = initial_orbital[1]
            coordinate_initial_second = initial_orbital[2]
            for coordinate in ["x", "y", "z"]:  # Determine non present coordinate
                if coordinate not in [coordinate_initial_first, coordinate_initial_second]:
                    coordinate_initial_third = coordinate
            if not special_orbital:
                coordinate_final_first = final_orbital[1]
                coordinate_final_second = final_orbital[2]
                [l_aux, m_aux, n_aux] = ([direction_cosines[coordinate_initial_first],
                                          direction_cosines[coordinate_initial_second],
                                          direction_cosines[coordinate_initial_third]])
                hopping = (3.*l_aux*m_aux*
                           direction_cosines[coordinate_final_first]*direction_cosines[coordinate_final_second]*Vdds)
                if initial_orbital == final_orbital:
                    hopping += (l_aux**2 + m_aux**2 - 4*l_aux**2*m_aux**2)*Vddp + (n_aux**2 + l_aux**2*m_aux**2)*Vddd
                else:
                    non_repeated_coordinates = list({coordinate_initial_first, coordinate_initial_second} ^
                                                    {coordinate_final_first, coordinate_final_second})
                    repeated_coordinate = list({coordinate_initial_first, coordinate_initial_second} &
                                               {coordinate_final_first, coordinate_final_second})[0]
                    m_aux = direction_cosines[repeated_coordinate]
                    [l_aux, n_aux] = [direction_cosines[non_repeated_coordinates[0]],
                                      direction_cosines[non_repeated_coordinates[1]]]
                    hopping += l_aux*n_aux*(1 - 4*m_aux**2)*Vddp + l_aux*n_aux*(m_aux**2 - 1)*Vddd
            elif final_orbital == "dx2-y2":
                hopping = (direction_cosines[coordinate_initial_first]*direction_cosines[coordinate_initial_second] *
                           (l*l - m*m)*(3./2*Vdds - 2*Vddp + Vddd/2.))
                if initial_orbital == "dyz":
                    hopping += (direction_cosines[coordinate_initial_first]*direction_cosines[coordinate_initial_second] *
                                (-Vddp + Vddd))
                elif initial_orbital == "dzx":
                    hopping += (direction_cosines[coordinate_initial_first]*direction_cosines[coordinate_initial_second] *
                                (Vddp - Vddd))
            elif final_orbital == "d3z2-r2":
                hopping = math.sqrt(3)*(direction_cosines[coordinate_initial_first] *
                                        direction_cosines[coordinate_initial_second]*(n*n - (l*l + m*m)/2))*Vdds
                if initial_orbital == 'dxy':
                    hopping += math.sqrt(3)*(-2*l*m*n*n*Vddp + (l*m*(1 + n*n)/2)*Vddd)
                else:
                    hopping += math.sqrt(3)*(direction_cosines[coordinate_initial_first]*
                    direction_cosines[coordinate_initial_second] * ((l*l + m*m - n*n)*Vddp - (l*l + m*m)/2.*Vddd))

        elif initial_orbital == "dx2-y2":
            hopping = 0.0
            if final_orbital == "dx2-y2":
                hopping += 3./4*(l*l - m*m)**2*Vdds + (l*l + m*m - (l*l - m*m)**2)*Vddp + (n*n + (l*l - m*m)**2/4)*Vddd
            elif final_orbital == "d3z2-r2":
                hopping += math.sqrt(3)*((l*l - m*m)*(n*n - (l*l + m*m)/2)*Vdds/2 + n*n*(m*m - l*l)*Vddp + ((1 + n*n) *
                              (l*l - m*m)/4)*Vddd)
        elif initial_orbital == "d3z2-r2":
            hopping = (n*n - (l*l + m*m)/2)**2*Vdds + 3*n*n*(l*l + m*m)*Vddp + 3./4*(l*l + m*m)**2*Vddd

        else:
            raise ValueError("Error: Shouldn't get called (some bug in orbital assertion in code)")

        return hopping


    def __initialize_spin_orbit_coupling(self) -> None:
        """ 
        Method to initialize the whole spin-orbit coupling matrix corresponding to the
        orbitals that participate in the tight-binding model.
        """

        dimension_block = 1 + 3 + 5
        # We hardcode the whole spin-orbit hamilonian up to d orbitals
        spin_orbit_hamiltonian = np.zeros([dimension_block*2, dimension_block*2], dtype=np.complex_)
        p_orbital_beginning = 1
        d_orbital_beginning = 4
        # p orbitals
        spin_orbit_hamiltonian[p_orbital_beginning, p_orbital_beginning + 1] = -1j
        spin_orbit_hamiltonian[p_orbital_beginning, dimension_block + p_orbital_beginning + 2] = 1
        spin_orbit_hamiltonian[p_orbital_beginning + 1, dimension_block + p_orbital_beginning + 2] = -1j
        spin_orbit_hamiltonian[p_orbital_beginning + 2, dimension_block + p_orbital_beginning] = -1
        spin_orbit_hamiltonian[p_orbital_beginning + 2, dimension_block + p_orbital_beginning + 1] = 1j
        spin_orbit_hamiltonian[dimension_block + p_orbital_beginning, dimension_block + p_orbital_beginning + 1] = 1j

        # d orbitals
        spin_orbit_hamiltonian[d_orbital_beginning, d_orbital_beginning + 3] = 2j
        spin_orbit_hamiltonian[d_orbital_beginning, dimension_block + d_orbital_beginning + 1] = 1
        spin_orbit_hamiltonian[d_orbital_beginning, dimension_block + d_orbital_beginning + 2] = -1j
        spin_orbit_hamiltonian[d_orbital_beginning + 1, d_orbital_beginning + 2] = 1j
        spin_orbit_hamiltonian[d_orbital_beginning + 1, dimension_block + d_orbital_beginning] = -1
        spin_orbit_hamiltonian[d_orbital_beginning + 1, dimension_block + d_orbital_beginning + 3] = -1j
        spin_orbit_hamiltonian[d_orbital_beginning + 1, dimension_block + d_orbital_beginning + 4] = -math.sqrt(3)*1j
        spin_orbit_hamiltonian[d_orbital_beginning + 2, dimension_block + d_orbital_beginning] = 1j
        spin_orbit_hamiltonian[d_orbital_beginning + 2, dimension_block + d_orbital_beginning + 3] = -1
        spin_orbit_hamiltonian[d_orbital_beginning + 2, dimension_block + d_orbital_beginning + 4] = math.sqrt(3)
        spin_orbit_hamiltonian[d_orbital_beginning + 3, dimension_block + d_orbital_beginning + 1] = 1j
        spin_orbit_hamiltonian[d_orbital_beginning + 3, dimension_block + d_orbital_beginning + 2] = 1
        spin_orbit_hamiltonian[d_orbital_beginning + 4, dimension_block + d_orbital_beginning + 1] = 1j*math.sqrt(3)
        spin_orbit_hamiltonian[d_orbital_beginning + 4, dimension_block + d_orbital_beginning + 2] = -math.sqrt(3)
        spin_orbit_hamiltonian[dimension_block + d_orbital_beginning, dimension_block + d_orbital_beginning + 3] = -2j
        spin_orbit_hamiltonian[dimension_block + d_orbital_beginning + 1,
                            dimension_block + d_orbital_beginning + 2] = -1j

        spin_orbit_hamiltonian += np.conj(spin_orbit_hamiltonian.T)

        self.spin_orbit_hamiltonian = spin_orbit_hamiltonian/3

    def __spin_orbit_h(self) -> None:
        """ 
        Method to obtain the actual spin-orbit hamiltonian that corresponds to the orbitals
        that participate in the model.
        """

        possible_orbitals = ['s', 'px', 'py', 'pz', 'dxy', 'dyz', 'dzx', 'dx2-y2', 'd3z2-r2']
        npossible_orbitals = len(possible_orbitals)
        spin_orbit_hamiltonian = sp.lil_matrix((self.basisdim, self.basisdim), dtype=np.complex_)
        matrix_index = 0
        for n, atom in enumerate(self.motif):
            species = int(atom[3])
            orbitals = self.configuration["Orbitals"][species]
            norbitals = self.norbitals[species]
            orbitals_indices = [index + i for index, orbital in enumerate(possible_orbitals) for i in [0, npossible_orbitals]
                                if orbital in orbitals]
            
            soc = (self.spin_orbit_hamiltonian[np.ix_(orbitals_indices, orbitals_indices)] *
                   self.configuration['Spin-orbit coupling'][species])
            spin_orbit_hamiltonian[matrix_index : matrix_index + norbitals, matrix_index : matrix_index + norbitals] = soc
            matrix_index += norbitals

        self.spin_orbit_hamiltonian = spin_orbit_hamiltonian

    def initialize_hamiltonian(self, find_bonds: bool = True, verbose: bool = True) -> None:
        """ 
        Routine to initialize the hamiltonian matrices which describe the system. 

        :find_bonds: Toggle on/or bond search before initializing the Hamiltonian. Useful
            when the bond search algorithm fails or we want to specify some bonds. Defaults to True.
        :verbose: Toggle print info. Defaults to True.
        """

        self.norbitals = [len(orbitals) for orbitals in self.configuration["Orbitals"]]
        self.basisdim = np.sum([self.norbitals[int(atom[3])] for atom in self.motif])

        if find_bonds:
            print('Computing first neighbours...\n')
            self.find_neighbours(mode=self._mode, nn=self.neighbours, r=self._r)
        self._determine_connected_unit_cells()

        hamiltonian = []
        for _ in self._unit_cell_list:
            hamiltonian.append(sp.lil_matrix((self.basisdim, self.basisdim), dtype=np.complex_))

        matrix_index = 0
        for n, atom in enumerate(self.motif):
            species = int(atom[3])  # To match list beginning on zero
            hamiltonian_atom_block = sp.diags(np.array(self.configuration['Onsite energy'][species]), format='lil')
            hamiltonian[0][matrix_index : matrix_index + self.norbitals[species],
                           matrix_index : matrix_index + self.norbitals[species]] = hamiltonian_atom_block
            matrix_index += self.norbitals[species]

        atom_index = {0: 0}
        for n, atom in enumerate(self.motif[1:]):
            atom_index[n + 1] = atom_index[n] + self.norbitals[int(self.motif[n][3])]

        for bond in self.bonds:
            initial_atom_index, final_atom_index, cell, neigh = bond
            initial_atom = self.motif[initial_atom_index][:3]
            initial_atom_species = int(self.motif[initial_atom_index][3])
            final_atom = self.motif[final_atom_index][:3]
            final_atom_species = int(self.motif[final_atom_index][3])
            iorbitals = self.configuration['Orbitals'][initial_atom_species]
            forbitals = self.configuration['Orbitals'][final_atom_species]
            for i, initial_orbital in enumerate(iorbitals):
                for j, final_orbital in enumerate(forbitals):
                    position_difference = np.array(final_atom) + np.array(cell) - np.array(initial_atom)
                    orbital_config = [initial_orbital, initial_atom_species, final_orbital, final_atom_species, neigh]
                    h_cell = self._unit_cell_list.index(list(cell))
                    hopping_amplitude = self._hopping_amplitude(position_difference, orbital_config)
                    hamiltonian[h_cell][atom_index[initial_atom_index] + i,
                                        atom_index[final_atom_index] + j] = hopping_amplitude
        if verbose:
            print("Hoppings computed")

        # Substract half-diagonal from all hamiltonian matrices to compensate transposing later
        #for h in hamiltonian:
        #    h -= sp.diags(h.diagonal()/2)

        # Check spinless or spinful model and initialize spin-orbit coupling
        if self.configuration['Spin']:
            self.norbitals = [2*norbitals for norbitals in self.norbitals]
            self.basisdim = self.basisdim * 2

            for index, cell in enumerate(self._unit_cell_list):
                hamiltonian[index] = sp.kron(hamiltonian[index], sp.eye(2, 2))

            # Add Zeeman term
            self.__zeeman_term(0)
            hamiltonian[0] += self.__zeeman

            if self.configuration['Spin-orbit coupling'] != 0:
                self.__initialize_spin_orbit_coupling()
                self.__spin_orbit_h()
                hamiltonian[0] += self.spin_orbit_hamiltonian
        
        if self.matrix_type == "dense":
            self.hamiltonian = [h.toarray() for h in hamiltonian]
        else:
            self.hamiltonian = [h.tocsr() for h in hamiltonian]

    def __zeeman_term(self, intensity: float) -> None:
        """ 
        Routine to incorporate a Zeeman term to the Hamiltonian. 
        
        :param intensity: Amplitude of the Zeeman term.
        """

        zeeman_h = sp.lil_matrix((self.basisdim, self.basisdim))
        matrix_index = 0
        for atom in self.motif:
            species = int(atom[3])
            zeeman_h[matrix_index: matrix_index + self.norbitals[species], 
                     matrix_index: matrix_index + self.norbitals[species]] = sp.kron(
                     np.array([[1, 0], [0, -1]]), 
                     sp.eye(self.norbitals[species]//2))
            matrix_index += self.norbitals[species]

        zeeman_h *= intensity
        self.__zeeman = zeeman_h

    def hamiltonian_k(self, k: Union[List, np.ndarray]) -> np.ndarray:
        """ 
        Routine to compute the Bloch Hamiltonian at a given k point. It can be given
        as a sparse matrix or a dense matrix depending on the matrix_type attribute.
        
        :param k: kpoint.
        :return: Bloch Hamiltonian evaluated at k.
        """

        if self.matrix_type == "dense":
            hamiltonian_k = np.zeros((self.basisdim, self.basisdim), dtype=np.complex_)
        else:
            hamiltonian_k = sp.csr_matrix((self.basisdim, self.basisdim), dtype=np.complex_)

        for cell_index, cell in enumerate(self._unit_cell_list):
            hamiltonian_k += self.hamiltonian[cell_index] * cmath.exp(1j*np.dot(k, cell))

        return hamiltonian_k
    
    def add_flat_band(self, energy: float, atom_index: int, spin: bool = False) -> SlaterKoster:
        """
        Adds a new orbital to one atom with given onsite energy and no hoppings to other atoms,
        resulting in a flat band in the Bloch Hamiltonian (localized state).
        Only works with crystalline systems.

        :param energy: Onsite energy of new orbital.
        :param atom_index: Index of atom to which we add the new orbital.
        :param spin: If set to True, two orbitals are added (one per spin).
        """

        if atom_index < 0:
            raise ValueError("Atom index must be a positive number.")
        elif atom_index > (self.natoms - 1):
            raise ValueError("Index of atom higher than number of chemical species (counting starts at 0).")
                
        index = 0
        for i in range(atom_index + 1):
            index += self.norbitals[i]

        nbasisdim = self.basisdim + 1 + spin
        for n, fock_matrix in enumerate(self.hamiltonian):
            new_fock_matrix = np.zeros([nbasisdim, nbasisdim], dtype=np.complex_)

            new_fock_matrix[0:index, 0:index] = fock_matrix[0:index, 0:index]
            new_fock_matrix[0:index, (index + 1 + spin):] = fock_matrix[0:index, index:]
            new_fock_matrix[(index + 1 + spin):, 0:index] = fock_matrix[index:, 0:index]
            new_fock_matrix[(index + 1 + spin):, (index + 1 + spin):] = fock_matrix[index:, index:]

            self.hamiltonian[n] = new_fock_matrix
        
        self.hamiltonian[0][index, index] = energy
        if spin:
            self.hamiltonian[0][index + 1, index + 1] = energy

        # Modify model attributes to take into account new orbital
        self.norbitals[atom_index] += 1 + spin
        self.basisdim += 1 + spin
        self.configuration["Filling"][atom_index] += 1 + spin

        return self
    
    def insert_SK_model(self, config: dict) -> SlaterKoster:
        """
        Takes a configuration dictionary of a system with only one atom and appends the corresponding
        model as a block to the current Hamiltonian. Note that it must describe the
        same lattice to make sense. Works only with crystalline systems.

        :param config: Configuration dictionary.
        """

        core_model = SlaterKoster(config)
        if len(core_model.motif) != 1:
            ValueError("New subsystem must have only one atom.")

        core_model.initialize_hamiltonian()
        nbasisdim = self.basisdim

        atom = core_model.motif[0][:3]
        self.motif = np.array(self.motif)
        atom_index = np.where(np.linalg.norm(self.motif[:, :3] - atom, axis=1) < 1E-4)[0]
        if len(atom_index) != 1:
            raise IndexError("Invalid atom positions in block SK model.")
        
        atom_index = atom_index[0]
        species = int(self.motif[atom_index, 3])
        
        index = 0
        for i in range(atom_index + 1):
            index += self.norbitals[i]
        
        norb = core_model.norbitals[0]
        nbasisdim += norb
        for n, fock_matrix in enumerate(self.hamiltonian):
            new_fock_matrix = np.zeros([nbasisdim, nbasisdim], dtype=np.complex_)

            new_fock_matrix[0:index, 0:index] = fock_matrix[0:index, 0:index]
            new_fock_matrix[0:index, (index + norb):] = fock_matrix[0:index, index:]
            new_fock_matrix[(index + norb):, 0:index] = fock_matrix[index:, 0:index]
            new_fock_matrix[(index + norb):, (index + norb):] = fock_matrix[index:, index:]

            new_fock_matrix[index:index + norb, index:index + norb] = core_model.hamiltonian[n]

            self.hamiltonian[n] = new_fock_matrix
        
        self.norbitals[atom_index] += core_model.norbitals[0]
        self.configuration["Filling"][species] += core_model.configuration["Filling"][0]
        self.basisdim = nbasisdim
        
        return self


    # ---------------------------- IO routines ----------------------------
    def export_model(self, filename: str, fmt: str = "%20.16f") -> None:
        """ 
        Routine to write the Hamiltonian matrices calculated to a file. 
        
        :param filename: Name of the file where we write the model.
        """

        h_fmt = [fmt+"%+"+fmt[1:]+"j" for _ in range(self.hamiltonian[0].shape[0])]
        with open(filename, "w") as file:
            # Write ndim, natoms, norbitals, ncells and bravais lattice basis vectors
            file.write("# dimension\n" + str(self.ndim) + "\n")
            file.write("# norbitals\n")
            np.savetxt(file, [self.norbitals], fmt="%i")
            file.write("# bravaislattice\n")
            np.savetxt(file, self.bravais_lattice, fmt=fmt)
            file.write("# motif\n")
            for atom in self.motif:
                np.savetxt(file, [atom], fmt=fmt)
            file.write("# bravaisvectors\n")
            for vector in self._unit_cell_list:
                np.savetxt(file, [vector], fmt=fmt)
            file.write("# hamiltonian\n")
            for h in self.hamiltonian:
                np.savetxt(file, h, fmt=h_fmt)
                file.write("&\n")
            if self.configuration["Filling"]:
                file.write("# filling\n" + str(self.filling) + "\n")
            file.write("#")


    @classmethod
    def import_model(cls, filename: str) -> 'SlaterKoster':
        """
        Routine to read the Hamiltonian matrix from a file.
        TODO: Has to be adapted to new file format. 

        :param filename: File with the configuration to load. Note that this
            file follows the format of export_model, not the one usually with fileparse.
        :return: Instance of SlaterKoster.
        """

        model = cls()
        file = open(filename, "r")
        file_contents = {}
        content = []
        for line in file.readlines():
            line = line.split()
            if line[0] == "#":
                arg = line[1]
                if not content:
                    continue
                file_contents[arg] = content
                content = []
            else:
                content.append(line)
        
        for key in file_contents.keys():
            if key == "dimension":
                model.dimension = int(file_contents[key][0])
            elif key == "norbitals":
                model.norbitals = [int(value) for value in file_contents[key][0]]
            elif key == "bravaislattice":
                bravais_lattice = []
                for vector in file_contents[key]:
                    bravais_lattice.append([float(value) for value in vector])


            # dimension
            line = file.readline()
            dimension = int(file.readline())

            # norbitals
            line = [int(value) for value in file.readline().split()]

            # bravaislattice
            

            ndim, natoms, norbitals, ncells = [int(num) for num in line]
            basisdim = norbitals * natoms
            bravais_lattice = []
            for i in range(ndim):
                line = file.readline().split()
                if len(line) != 3:
                    print("Unexpected line found, exiting...")
                    sys.exit(1)
                bravais_lattice.append([float(num) for num in line])

            motif = []
            for i in range(natoms):
                line = file.readline().split()
                if len(line) != 3:
                    print("Unexpected line found, exiting...")
                    sys.exit(1)
                motif.append([float(num) for num in line])

            unit_cell_list = []
            hamiltonian = []
            hamiltonian_matrix = np.zeros([basisdim, basisdim], dtype=np.complex_)
            for line in file.readlines():
                line = line.split()
                if len(line) == 3:
                    unit_cell_list.append([float(num) for num in line])
                elif line[0] == "#":
                    it = 0
                    hamiltonian.append(hamiltonian_matrix)
                    hamiltonian_matrix = np.zeros([basisdim, basisdim], dtype=np.complex_)
                else:
                    hamiltonian_matrix[it, :] = [complex(line[2*i], line[2*i + 1]) for i in range(len(line))]
                    it += 1

        if ncells != len(unit_cell_list):
            print("Mismatch between number of Bloch matrices provided and declared, exiting...")
            sys.exit(1)

        model = cls()
        model.system_name = filename
        model.ndim = ndim
        model.norbitals = norbitals
        model.natoms = natoms
        model.basisdim = basisdim
        model._unit_cell_list = unit_cell_list
        model.bravais_lattice = bravais_lattice
        model.hamiltonian = hamiltonian

        return model


class AmorphousSlaterKoster(SlaterKoster):
    """ 
    Extension of the Slater-Koster tight binding model to describe amorphous solids,
    taking the crystalline solid as the reference point. This model by default works in radius mode. 

    :param configuration: Configuration dictionary as generated by parsefile.
    :param r: Radius up to which bonds are considered. Defaults to None.
    :param boundary: 'PBC' or 'OBC'. Defaults to 'PBC'.
    """

    def __init__(self, configuration=None, r=None, boundary="PBC"):
        super().__init__(configuration, mode='radius', r=r, boundary=boundary)

        # Specific attributes of AmorphousSlaterKoster
        self._reference_lengths = {}
        self._decay_amplitude   = None
        self._decay_mode        = 'exponential'
        self._mode              = 'radius'

    # Properties
    @property
    def reference_lengths(self):
        return self._reference_lengths

    @reference_lengths.setter
    def reference_lengths(self, *values):
        print('reference_lengths must be mutated using the set_reference_length method')

    @property
    def decay_amplitude(self):
        return self._decay_amplitude
    
    @decay_amplitude.setter
    def decay_amplitude(self, amplitude):
        if amplitude < 0:
            raise ValueError('decay_amplitude: amplitude must be positive number')
        self._decay_amplitude = amplitude
    
    @property
    def decay_mode(self):
        return self._decay_mode
    
    @decay_mode.setter
    def decay_mode(self, mode):
        if mode != "exponential" and mode != "polynomial":
            raise ValueError('decay_mode: mode must be either exponential or polynomial')
        self._decay_mode = mode

    def set_reference_length(self, first_species: int, second_species: int, length: float) -> None:
        """ 
        Method to set manually the reference bond length between two species. 
        
        :param first_species: Index of first species.
        :param second_species: Index of second species.
        :param length: Reference length value.
        """

        if first_species > self.species or second_species > self.species:
            raise ValueError('set_reference_length: provided species must be numbers between 0 and nspecies - 1')
        if first_species > second_species:
            species_pair = str(second_species) + str(first_species)
        else:
            species_pair = str(first_species) + str(second_species)
        self._reference_lengths[species_pair] = length
        print(f"Reference lengths for species: {self._reference_lengths}")

    def __compute_reference_lengths(self):
        """ 
        Private method to compute the length between different chemical species in the crystalline solid. 
        To be used only when the model is initialized with a configuration file corresponding to a crystalline solid. 
        """

        # First compute neighbours in crystalline solid
        print("Computing reference lengths...")
        self.find_neighbours(mode=self._mode, r = self._r)
        length_dict = {}
        motif = np.array(self.motif)
        for i, j, cell, _ in self.bonds:
            initial_species = int(self.motif[i][3])
            final_species   = int(self.motif[j][3])
            distance = np.linalg.norm(motif[i, :3] - motif[j, :3] - cell)
            species_pair = str(final_species) + str(initial_species) if (i > j) else str(initial_species) + str(final_species)
            if species_pair in length_dict:
                if distance < length_dict[species_pair]:
                    length_dict[species_pair] = distance
            else:
                length_dict[species_pair] = distance

        print(f"Reference lengths for species: {length_dict}")
        self._reference_lengths = length_dict

    @overrides(SlaterKoster)
    def _hopping_amplitude(self, position_diff: np.ndarray, orbitals: list) -> float:
        """ 
        Method to incorporate a decay to the hoppings as computed within a Slater-Koster model to describe
        variable length bonds 
        
        :param position_diff: Vector connecting the initial and the final atoms.
        :param orbitals: List [initial_orbital, initial_species, final_orbital, final_species, neigh_index]
        :return: Hopping amplitude.
        """

        initial_species = int(orbitals[1])
        final_species   = int(orbitals[3])
        species_pair = str(initial_species) + str(final_species)
        if species_pair not in self.reference_lengths.keys():
            species_pair = str(final_species) + str(initial_species)
        
        reference_bond_length = self.reference_lengths[species_pair]
        r = np.linalg.norm(position_diff)
        displacement = round(r - reference_bond_length, 4)
        
        hopping = super()._hopping_amplitude(position_diff, orbitals)
        
        if self.decay_mode == "exponential":
            hopping *= np.exp(-self.decay_amplitude*displacement)
        else:
            # hopping *= 1./(1 + self.decay_amplitude*(r - reference_bond_length))
            hopping *= (r/reference_bond_length)**(-self.decay_amplitude)

        return hopping

    @overrides(SlaterKoster)
    def initialize_hamiltonian(self, find_bonds: bool = True, override_bond_lengths: bool = False) -> None:
        """
        Method to initialize the 

        :param find_bonds: Toggle on/off bond search. Defaults to True.
        :param override_bond_length: Overwrite previous reference bond lengths.
        """

        if override_bond_lengths:
            self.__compute_reference_lengths()
        super().initialize_hamiltonian(find_bonds=find_bonds)


class BHZ(System, FrozenClass):
    """
    Implementation of generalized BHZ model. The model is based on a
    2D square lattice of side a=1, and is a four band model. The model takes three parameters:
    
    :param g: Mixing term
    :param u: Sublattice potential
    :param c: Coupling operator "amplitude"
    """

    def __init__(self, g, u, c):
        super().__init__(system_name="BHZ model",
                         crystal=Crystal([[1, 0, 0], [0, 1, 0]],
                                         motif=[[0, 0, 0, 0]]))
        self.norbitals = [4]
        self.filling = 2
        self.basisdim = self.norbitals[0]
        self.g = g
        self.u = u
        self.c = c

        self._freeze()

    def hamiltonian_k(self, k: Union[List, np.ndarray]) -> np.ndarray:
        """
        Method to compute the Bloch Hamiltonian.
        
        :param k: Array used to compute the Bloch Hamiltonian.
        :return: H(k).
        """
        
        global sigma_x, sigma_y, sigma_z
        coupling = self.c * sigma_y

        id2 = np.eye(2)
        kx, ky = k[0], k[1]

        hamiltonian = (np.kron(id2, (self.u + np.cos(kx) + np.cos(ky)) * sigma_z + np.sin(ky) * sigma_y) +
                       np.kron(sigma_z, np.sin(kx) * sigma_x) + np.kron(sigma_x, coupling) +
                       self.g * np.kron(sigma_z, sigma_y) * (np.cos(kx) + np.cos(7 * ky) - 2))

        return hamiltonian


class WilsonFermions2D(System, FrozenClass):
    """
    Implementation of Wilson-fermions model. This model takes a 2D square lattice of side a, it is
    a four band model.

    :param side: Length of lattice (square) side. Defaults to 1.
    :param t: Hopping amplitude. Defaults tot 1.
    :param m: Mass term. Defaults to 1.
    """

    def __init__(self, side=1, t=1, m=1):
        super().__init__(system_name="Wilson-fermions 2D",
                         crystal=Crystal([[side, 0, 0], [0, side, 0]],
                                         motif=[[0, 0, 0, 0]]))
        self.system_name = "Wilson-fermions model"
        self.norbitals = 4
        self.basisdim = self.norbitals * 1
        self.filling = 0.5

        self.a = side
        self.t = t
        self.m = m

        self._freeze()

    def hamiltonian_k(self, k: Union[List, np.ndarray]) -> np.ndarray:
        """
        Method to compute the Bloch Hamiltonian.
        
        :param k: Array used to compute the Bloch Hamiltonian.
        :return: H(k).
        """

        global sigma_x, sigma_y, sigma_z

        alpha_x = np.kron(sigma_x, sigma_x)
        alpha_y = np.kron(sigma_x, sigma_y)
        beta = np.kron(sigma_z, np.eye(2, dtype=np.complex_))

        hamiltonian = self.t * (np.sin(k[0] * self.a) * alpha_x + np.sin(k[1] * self.a) * alpha_y) + (
                np.cos(k[0] * self.a) + np.cos(k[1] * self.a) + self.m - 3) * beta

        return hamiltonian


class WilsonFermions3D(System, FrozenClass):
    """
    Implementation of Wilson-fermions model. This model takes a 3D square lattice of side a, it is
    a four band model.

    :param side: Length of lattice (square) side. Defaults to 1.
    :param t: Hopping amplitude. Defaults tot 1.
    :param m: Mass term. Defaults to 1.
    """

    def __init__(self, side=1, t=1, m=1):
        super().__init__(system_name="Wilson-fermions 3D",
                         crystal=Crystal([[side, 0, 0], [0, side, 0], [0, 0, side]],
                                         motif=[[0, 0, 0, 0]]))
        self.num_orbitals = 4
        self.basisdim = self.num_orbitals * self.natoms

        self.a = side
        self.t = t
        self.m = m

        self._freeze()

    def hamiltonian_k(self, k: Union[List, np.ndarray]) -> np.ndarray:
        """
        Method to compute the Bloch Hamiltonian.
        
        :param k: Array used to compute the Bloch Hamiltonian.
        :return: H(k).
        """

        global sigma_x, sigma_y, sigma_z

        alpha_x = np.kron(sigma_x, sigma_x)
        alpha_y = np.kron(sigma_x, sigma_y)
        alpha_z = np.kron(sigma_x, sigma_z)
        beta = np.kron(sigma_z, np.eye(2, dtype=np.complex_))

        hamiltonian = (self.t * (np.sin(k[0] * self.a) * alpha_x + np.sin(k[1] * self.a) * alpha_y +
                                 np.sin(k[2] * self.a) * alpha_z) +
                       (np.cos(k[0] * self.a) + np.cos(k[1] * self.a) + np.cos(k[2] * self.a) +
                        self.m - 3) * beta)

        return hamiltonian


class WilsonAmorphous(System):
    """
    Implementation of 3D Wilson-fermion model in real-space. This models is a generalization of the
    standard Wilson-fermion model, which uses a cubic unit cell. Instead, this allows for any spatial
    distribution of the atoms, and since the Hamiltonian is built in real-space any number of atoms
    can be considered.

    :param side: First-neighbour distance in the crystalline system. Defaults to 1.
    :param t: Hopping parameter. Defaults to 1.
    :param m: Mass term. Defaults to 1.
    :param r: Cutoff distance for neighbour interaction. Defaults to 1.
    """

    def __init__(self, side=1, t=1, m=1, r=1.1):
        super().__init__(system_name="Wilson Amorphous model",
                         crystal=Crystal([[side, 0, 0], [0, side, 0], [0, 0, side]],
                                         motif=[[0, 0, 0, 0]]))

        self.filling = 2
        self.norbitals = [4]
        self.basisdim = self.norbitals[0] * len(self.motif)
        self.boundary = "PBC"

        self.a = side
        self.t = t
        self.m = m
        self.r = r
        self.parameters = {"C": self.t, "a": self.a, "M": self.m}

    @staticmethod
    def _hopping_matrix(initial_position: np.ndarray, final_position: np.ndarray, parameters: dict) -> np.ndarray:
        """ 
        Computes hopping matrix according to Wilson-fermion model for two
        given atomic positions.

        :param initial_position: Array 1x3 initial position.
        :param final_position: Array 1x3 final position.
        :param parameters: Dictionary{C, a, M}
        :return: Hopping matrix.
        """

        x, y, z = final_position - initial_position
        r = math.sqrt(x ** 2 + y ** 2 + z ** 2)
        phi = math.atan2(y, x)
        theta = math.acos(z / r)

        hopping = np.zeros([4, 4], dtype=np.complex_)
        np.fill_diagonal(hopping, [1, 1, -1, -1])
        offdiag_block = np.array([[-1j * math.cos(theta), -1j * cmath.exp(-1j * phi) * math.sin(theta)],
                                  [-1j * cmath.exp(1j * phi) * math.sin(theta), 1j * math.cos(theta)]],
                                 dtype=np.complex_)
        hopping[0:2, 2:4] = offdiag_block
        hopping[2:4, 0:2] = offdiag_block
        hopping *= 0.5
        hopping *= parameters["C"] * math.exp(1 - r / parameters["a"])

        return hopping

    def initialize_hamiltonian(self, find_bonds: bool = True) -> None:
        """ 
        Routine to initialize the matrices that compose the Bloch Hamiltonian
        
        :param find_bonds: Toggle to find bonds in the system before initializating
            the Hamiltonian.
        """

        self.basisdim = self.natoms * self.norbitals[0]
        norbitals = 4
        if find_bonds:
            print("Computing neighbours...")
            self.find_neighbours(mode="radius", r=self.r)
        if not find_bonds and self.bonds is None:
            raise ArithmeticError("No bonds found to initialize Hamiltonian, exiting...")
        self._determine_connected_unit_cells()

        hamiltonian = []
        for _ in self._unit_cell_list:
            hamiltonian.append(np.zeros([self.basisdim, self.basisdim], dtype=np.complex_))

        hamiltonian_atom_block = np.diag(np.array([-3 + self.m, -3 + self.m,
                                                   3 - self.m, 3 - self.m])*0.5)
        for n, atom in enumerate(self.motif):
            hamiltonian[0][norbitals * n:norbitals * (n + 1),
                           norbitals * n:norbitals * (n + 1)] = hamiltonian_atom_block

        for bond in self.bonds:
            initial_atom_index, final_atom_index, cell, _ = bond
            initial_atom = self.motif[initial_atom_index][:3]
            final_atom = self.motif[final_atom_index][:3]
            neigh_position = np.array(final_atom) + np.array(cell)
            h_cell = self._unit_cell_list.index(list(cell))
            hamiltonian[h_cell][4 * initial_atom_index:4 * (initial_atom_index + 1),
                                4 * final_atom_index:4 * (final_atom_index + 1)] = \
                self._hopping_matrix(initial_atom, neigh_position, self.parameters)

        self.hamiltonian = hamiltonian

    def hamiltonian_k(self, k: Union[list, np.ndarray]) -> np.ndarray:
        """
        Routine to evaluate the Bloch Hamiltonian at a given k point. It adds the k dependency of the Bloch Hamiltonian
        through the complex exponentials.

        :param k: k vector (Array 1x3)
        :return: Bloch Hamiltonian matrix
        """

        dimension = len(self.hamiltonian[0])
        hamiltonian_k = np.zeros((dimension, dimension), dtype=np.complex_)
        for cell_index, cell in enumerate(self._unit_cell_list):
            hamiltonian_k += (self.hamiltonian[cell_index] * cmath.exp(1j * np.dot(k, cell)))

        hamiltonian_k = (hamiltonian_k + np.transpose(np.conjugate(hamiltonian_k)))
        return hamiltonian_k
    

class HaldaneModel(System, FrozenClass):
    """
    Implementation of Haldane model in reciprocal space.

    :param t1: Strength of first-neighbour hoppings.
    :param t2: Strength of second-neighbour hoppings.
    :param phi: Phase of time-reversal breaking term.
    :param m: Mass term.
    """

    def __init__(self, t1: float = -1, t2: float = 0, phi: float = 0, m: float = 0):
        
        lattice_vectors = np.array([[np.sqrt(3)/2,  1./2, 0], 
                                    [np.sqrt(3)/2, -1./2, 0]])
        motif = np.array([[            0, 0, 0, 0],
                          [1./np.sqrt(3), 0, 0, 0]])
        super().__init__(system_name="Haldane model", crystal=Crystal(lattice_vectors, motif))

        self.t1  = t1
        self.t2  = t2
        self.phi = phi
        self.m   = m 

        

        self.norbitals = [1]
        self.basisdim  = self.norbitals[0] * len(self.motif)
        self.boundary  = "PBC"
        self.filling   = 1

        self._freeze()

    def hamiltonian_k(self, k: Union[List, np.ndarray]) -> np.ndarray:
        """
        Routine to evaluate the Bloch Hamiltonian at a given k point. It adds the k dependency of the Bloch Hamiltonian
        through the complex exponentials.

        :param k: k vector (Array 1x3)
        :return: Bloch Hamiltonian matrix
        """

        global sigma_x, sigma_y, sigma_z
        identity = np.eye(2)

        a = 1./np.sqrt(3)
        angle = 60 * np.pi / 180
        a1 = np.array([a, 0., 0.]) 
        a2 = np.array([-a * np.cos(angle), a * np.sin(angle), 0]) 
        a3 = np.array([-a * np.cos(angle), -a * np.sin(angle), 0])

        first_neighbours = [a1, a2, a3]
        second_neighbours = [-a2 + a3, -a3 + a1, -a1 + a2]

        h = np.zeros([2, 2], dtype=np.complex_)

        h += 2 * self.t2 * np.cos(self.phi) * np.sum([np.cos(np.dot(k, b)) for b in second_neighbours]) * identity
        h += self.t1 * np.sum([np.cos(np.dot(k, a)) for a in first_neighbours]) * sigma_x
        h += self.t1 * np.sum([np.sin(np.dot(k, a)) for a in first_neighbours]) * sigma_y
        h += (self.m - 2 * self.t2 * np.sin(self.phi) * np.sum([np.sin(np.dot(k, b)) for b in second_neighbours])) * sigma_z
        
        return h
    

class AgarwalaChern(System):
    """
    Implementation of the Agarwala model of amorphous Chern insulators in 2D, which is a generalization 
    of the BHZ model for arbitrary atomic positions, with additional terms to break particle-hole and inversion symmetry.

    :param side: First-neighbour distance in the crystalline system. Defaults to 1.
    :param t: Hopping parameter. Defaults to 1.
    :param m: Mass term. Defaults to 1.
    :param t2: Particle-hole symmetry breaking parameter. Defaults to 0.
    :param lda: Inversion symmetry breaking parameters. Defaults to 0
    :param r: Cutoff distance for neighbour interaction. Defaults to side*1.1.
    """

    def __init__(self, side=1, t=1, m=1, t2=0, lda=0,  r=1.1):
        super().__init__(system_name="Wilson Amorphous model",
                         crystal=Crystal([[side, 0, 0], [0, side, 0]],
                                         motif=[[0, 0, 0, 0]]))

        self.filling = 1
        self.norbitals = [2]
        self.basisdim = self.norbitals[0] * len(self.motif)
        self.boundary = "PBC"

        self.a = side
        self.t = t
        self.m = m
        self.t2 = t2
        self.lda = lda
        self.r = r * self.a
        self.parameters = {"C": self.t, "a": self.a, "M": self.m, "t2": self.t2, "lambda": lda}

    def _hopping_matrix(self, initial_position: np.ndarray, final_position: np.ndarray) -> np.ndarray:
        """ 
        Computes hopping matrix according to Wilson-fermion model for two
        given atomic positions.

        :param initial_position: Array 1x3 initial position.
        :param final_position: Array 1x3 final position.
        :return: Hopping matrix.
        """

        x, y, z = final_position - initial_position
        r = math.sqrt(x ** 2 + y ** 2 + z ** 2)
        theta = math.atan2(y, x)

        hopping = np.zeros([2, 2], dtype=np.complex_)
        hopping[0, 0] = -1 + self.t2
        hopping[1, 1] =  1 + self.t2
        hopping[0, 1] = -1j*cmath.exp(-1j*theta) + self.lda * ((1 + 1j)*math.sin(theta)**2 - 1)
        hopping[1, 0] = -1j*cmath.exp(1j*theta) + self.lda * ((1 - 1j)*math.sin(theta)**2 - 1)
        
        hopping *= 0.5 * math.exp(-(r - self.a))

        return hopping

    def initialize_hamiltonian(self, find_bonds: bool = True) -> None:
        """ 
        Routine to initialize the matrices that compose the Bloch Hamiltonian
        
        :param find_bonds: Toggle to find bonds in the system before initializating
            the Hamiltonian.
        """

        self.basisdim = self.natoms * self.norbitals[0]
        if find_bonds:
            print("Computing neighbours...")
            self.find_neighbours(mode="radius", r=self.r)
        if not find_bonds and self.bonds is None:
            raise ArithmeticError("No bonds found to initialize Hamiltonian, exiting...")
        self._determine_connected_unit_cells()

        hamiltonian = []
        for _ in self._unit_cell_list:
            hamiltonian.append(np.zeros([self.basisdim, self.basisdim], dtype=np.complex_))

        hamiltonian_atom_block = np.array([[       2 + self.m, (1 - 1j)*self.lda], 
                                           [(1 + 1j)*self.lda,       -2 - self.m]])

        for n, atom in enumerate(self.motif):
            hamiltonian[0][self.norbitals[0] * n:self.norbitals[0] * (n + 1),
                           self.norbitals[0] * n:self.norbitals[0] * (n + 1)] = hamiltonian_atom_block

        for bond in self.bonds:
            initial_atom_index, final_atom_index, cell, _ = bond
            initial_atom = self.motif[initial_atom_index][:3]
            final_atom = self.motif[final_atom_index][:3]
            neigh_position = np.array(final_atom) + np.array(cell)
            h_cell = self._unit_cell_list.index(list(cell))
            hamiltonian[h_cell][2 * initial_atom_index:2 * (initial_atom_index + 1),
                                2 * final_atom_index:2 * (final_atom_index + 1)] = \
                self._hopping_matrix(initial_atom, neigh_position)

        self.hamiltonian = hamiltonian

    def hamiltonian_k(self, k: Union[list, np.ndarray]) -> np.ndarray:
        """
        Routine to evaluate the Bloch Hamiltonian at a given k point. It adds the k dependency of the Bloch Hamiltonian
        through the complex exponentials.

        :param k: k vector (Array 1x3)
        :param conditions: defaults to PBC. Can be either PBC or OBC
        :return: Bloch Hamiltonian matrix
        """

        dimension = len(self.hamiltonian[0])
        hamiltonian_k = np.zeros((dimension, dimension), dtype=np.complex_)
        for cell_index, cell in enumerate(self._unit_cell_list):
            hamiltonian_k += (self.hamiltonian[cell_index] * cmath.exp(1j * np.dot(k, cell)))

        return hamiltonian_k


class RealSpace(System):
    """ 
    Class to construct toy models, in which one sets the hoppings manually. This models
    are by default OBC; by setting one hopping between different unit cells it automatically becomes
    PBC. 
    
    :param system_name: Name of system. Defaults to None.
    :param bravais_lattice: Bravais basis.
    :param motif: List of atoms of the solid.
    """
    
    def __init__(self, system_name: str = None, bravais_lattice: np.ndarray = None, motif: np.ndarray = None):
        super().__init__(system_name=system_name, bravais_lattice=bravais_lattice, motif=motif)
        self.onsite_energy = None
        self.hoppings = {}
        self.boundary = "OBC"
        self.orbitals = None
        self.bond_graphs = None

    def add_onsite_energy(self, index: int, energy: float) -> None:
        """ 
        Method to specify the onsite energy of a given atom. 
        
        :param index: Index of atom.
        :param energy: Value of onsite energy.
        """
        
        self.hoppings[index] = energy

    def add_onsite_energies(self, indices: List[float], energies: List[float]) -> None:
        """
        Same as add_onsite_energy. Used to specify the onsite energies of a collection
        of atom indices. 
        
        :param indices: List with indices of atoms.
        :param energies: List with onsite energies.
        """

        for index, energy in zip(indices, energies):
            self.add_onsite_energy(index, energy)

    def add_hopping(self, hopping: complex, initial: int, final: int, cell: list = [0., 0., 0.]) -> None:
        """ 
        Method to add a hopping between two atoms of the motif.
        NB: The hopping has a specified direction, from initial to final. Since the final
        Hamiltonian is computed taking into account hermiticity, it is not necessary to specify the hopping
        in the other direction.
        Parameters:
        :param hopping: Value of hopping.
        :param initial: Index of first atom of bond.
        :param final: Index of final atom of bond.
        :param cell: Bravais vector connecting the cells of the two atoms. Defaults to zero. 
        """

        assert type(initial) == int, "initial must be an integer"
        assert type(final)   == int, "initial must be an integer"
        assert type(hopping) != str, "hopping must be a complex number"

        self.add_bond(initial, final, cell)
        bond_index = len(self.bonds) - 1
        self.hoppings.append([bond_index, hopping])

    def add_hoppings(self, hoppings: List[complex], initial: List[int], final: List[int], cells: List[list] = None) -> None:
        """ 
        Same method as add_hopping but we input a list of hoppings at once.
        Parameters:
        :param hoppings: List with values of hoppings.
        :param initial: List of indices of initial atom of the hoppings.
        :param final: List of indices of final atoms of the motif.
        :param cells: Each row denotes the Bravais vector connecting two cells. Defaults to None. 
        """
        
        if len(hoppings) != len(initial) or len(initial) != len(final):
            raise ValueError("Provided list must have the same length")

        if cells is None:
            cells = np.zeros([len(hoppings), 3])
        else:
            self.boundary = "PBC"
            cells = np.array(cells)
        for n, hopping in enumerate(hoppings):
            self.add_hopping(hopping, initial[n], final[n], cells[n, :])

    def specify_hopping_amplitude(self, amplitude: float, bond_index: int) -> None:
        """ 
        Instead of adding the hopping directly as a bond with an amplitude, we
        can also specify the hopping amplitude of a preexisting bond. 
        
        :param amplitude: Hopping amplitude.
        :param bond_index: Index of bond.
        """

        self.hoppings.append([bond_index, amplitude])

    def specify_hopping_amplitudes(self, amplitudes: List[float], bond_indices: List[int]) -> None:
        """ 
        Same as specify_hopping_amplitude but for a list of amplitudes-bond indices. 
        
        :param amplitudes: List of amplitudes of bonds.
        :param bond_indices: List of indices of bonds.
        """
        
        
        for amplitude, bond_index in zip(amplitudes, bond_indices):
            self.specify_hopping_amplitude(amplitude, bond_index)

    def __construct_bond_graph(self) -> None:
        """ 
        Private method to create the graph of bonds between all atoms 
        """
        
        bond_graphs = []
        for _ in self._unit_cell_list:
            bond_graphs.append(np.zeros([self.natoms, self.natoms]))

        for bond, _ in self.hoppings:
            initial_atom, final_atom, cell = bond
            cell_index = self._unit_cell_list.index(cell)
            bond_graphs[cell_index][initial_atom, final_atom] += 1

        for i in range(self._unit_cell_list):
            bond_graphs[i] += bond_graphs[i].T

        self.bond_graphs = bond_graphs

    def __calculate_orbitals(self) -> None:
        """ 
        Private method to compute the dimension of the basis for the model, taking into account
        the number of bonds present for each atom of the motif. 
        """
        
        orbitals = []
        for i in range(self.natoms):
            max_orbitals = 0
            for j in range(len(self._unit_cell_list)):
                max_bonds_in_unit_cell = np.max(self.bond_graphs[j][i, :])
                if max_bonds_in_unit_cell > max_orbitals:
                    max_orbitals = max_bonds_in_unit_cell
            orbitals.append(max_orbitals)

        orbitals[orbitals == 0] = 1
        self.orbitals = orbitals

    def __calculate_basisdim(self) -> None:
        """ 
        Method to compute the dimension of the basis based on the orbitals of each atom 
        """
        
        self.basisdim = np.sum(self.orbitals)

    def initialize_hamiltonian(self) -> None:
        """
        Method to set up the matrices that compose the Hamiltonian, either the Bloch Hamiltonian
        or the real-space one. 
        TODO initialize_hamiltonian RSmodel
        """

        # First call private methods to compute orbitals per atom and basis dimension
        self._determine_connected_unit_cells()
        self.__calculate_orbitals()
        self.__calculate_basisdim()

        # Build the matrices that compose the Bloch Hamiltonian
        assert len(self.bonds) == len(self.hoppings)
        hamiltonian = []
        for _ in range(self._unit_cell_list):
            hamiltonian.append(np.zeros((self.basisdim, self.basisdim), dtype=np.complex))

        # Onsite energies
        for n in range(len(self.natoms)):
            hamiltonian[0][n:n + self.orbitals[n], n:n+self.orbitals[n]] = self.onsite_energy[n]/2.

        # Hoppings
        for bond_index, amplitude in self.hoppings:
            bond = self.bonds[bond_index]
            initial_atom_index, final_atom_index, cell = bond

    def hamiltonian_k(self, k):
        hamiltonian_k = np.zeros((self.basisdim, self.basisdim), dtype=np.complex_)
        for cell_index, cell in enumerate(self._unit_cell_list):
            hamiltonian_k += (self.hamiltonian[cell_index] * cmath.exp(1j * np.dot(k, cell)))

        return hamiltonian_k


class Stack(System):
    """ 
    Class that implements stacking of bidimensional systems. Its main task is stacking
    correctly the different layers and adding the hoppings between layers. The hamiltonian within
    each layer is delegated upon the class responsible of creating given system. 

    :param height: Distance between layers.
    :param bottom_layer: System corresponding to the bottom layer.
    :param top_layer: System corresponding to the top layer. Defaults to None.
    """

    def __init__(self, height, bottom_layer: System, top_layer: System = None) -> None:
        if bottom_layer.ndim != 2 or (top_layer and top_layer.ndim != 2):
            raise ValueError("Systems must be two-dimensional")
        if height < 0:
            raise ValueError("Interlayer height must be a positive number")

        if top_layer:
            for vector_bottom, vector_top in zip(bottom_layer.bravais_lattice, top_layer.bravais_lattice):
                if np.linalg.norm(vector_bottom - vector_top) > 1E-2:
                    raise ValueError('Both systems must have the same Bravais lattice')
        else:
            top_layer = bottom_layer

        super().__init__()
        self.ndim         = 2
        self.bottom_layer = bottom_layer
        print(self.bottom_layer)
        self.top_layer    = top_layer
        self.height       = height
        self.vertical = False if top_layer else True
        self.__interlayer_hopping = 0
        self.initialize_system_attributes()  

    @property
    def filling(self):
        nelectrons = self.bottom_layer.filling + self.top_layer.filling
        return nelectrons

    @property
    def interlayer_hopping(self):
        return self.__interlayer_hopping

    @interlayer_hopping.setter
    def interlayer_hopping(self, hopping: int):
        """ 
        Setter for interlayer hopping. Note that it is a single value instead
        of multiple values for different orbitals as in SK models. 
        """
        
        self.__interlayer_hopping = hopping

    def initialize_system_attributes(self) -> None:
        """ 
        Routine to initialize all System atttributes from the bottom layer.
        """

        # Bravais lattice
        self.bravais_lattice = self.bottom_layer.bravais_lattice

        # Motif
        top_layer_motif = np.copy(self.top_layer.motif)
        top_layer_motif[:, 2] += np.max(np.array(self.bottom_layer.motif)[:, 2]) + self.height
        self.motif  = np.concatenate((self.bottom_layer.motif, top_layer_motif), axis=0)
        self.natoms = len(self.motif)

    def __find_closest_atom_from_next_layer(self, atom_index: int) -> int:
        """ 
        Routine to find closest atom from next layer. 
        
        :param atom_index: Index of atom from bottom layer.
        :return: Closest atom from top layer.
        """

        atom_position = self.bottom_layer.motif[atom_index][:3]
        distances = np.linalg.norm(np.array(self.top_layer.motif)[:, :3] + self.height - atom_position, axis=1)
        closest_atom_index = np.argmin(distances)

        return closest_atom_index

    def find_interlayer_bonds(self, vertical: bool = False) -> list:
        """ 
        Routine to find list of all bonds between layers. If vertical, bonds are
        created directly without search. 

        :param vertical: If True, the bond is taken simply as the same atom but from the
            top layer, i.e. it is for two equal layers. Defaults to False.
        :return: List of interlayer bonds.
        """

        interlayer_bonds = []
        origin = np.array([0., 0., 0.])
        for atom_index in range(self.bottom_layer.natoms):
            if vertical:
                next_layer_atom_index = atom_index
            else:
                next_layer_atom_index = self.__find_closest_atom_from_next_layer(atom_index)
            interlayer_bonds.append([atom_index, next_layer_atom_index + self.bottom_layer.natoms, origin, 1])
        
        return interlayer_bonds
    
    def initialize_hamiltonian(self):
        """ 
        Init. hamiltonian method for Stack class. 
        """ 

        # Check if interlayer hopping was specified
        if self.interlayer_hopping == 0:
            print('Warning: Interlayer hopping is zero')

        # First initialize hamiltonian of each layer and bonds between layers
        self.bottom_layer.initialize_hamiltonian()
        self.top_layer.initialize_hamiltonian()
        self._unit_cell_list = self.bottom_layer._unit_cell_list
        self.interlayer_bonds = self.find_interlayer_bonds(self.vertical)
        
        # Works only for
        self.norbitals = self.bottom_layer.norbitals

        # Compute basisdim
        self.basisdim = self.bottom_layer.basisdim + self.top_layer.basisdim

        # Now glue the hamiltonians of each layer together
        self.hamiltonian = []
        for cell_index in range(len(self.bottom_layer._unit_cell_list)):
            h_cell = np.zeros([self.basisdim, self.basisdim], dtype=np.complex_)
            h_cell[:self.bottom_layer.basisdim, 
                   :self.bottom_layer.basisdim] = self.bottom_layer.hamiltonian[cell_index]
            h_cell[self.bottom_layer.basisdim:self.basisdim, 
                   self.bottom_layer.basisdim:self.basisdim] = self.top_layer.hamiltonian[cell_index]
            
            self.hamiltonian.append(h_cell)

        # Precompute dictionary with positions of each atom within the hamiltonian
        bottom_atom_index_dictionary, top_atom_index_dictionary = {}, {}
        counter = 0
        for index, atom in enumerate(self.bottom_layer.motif):
            bottom_atom_index_dictionary[index] = counter
            counter += self.bottom_layer.norbitals[int(atom[3])]
        counter = 0
        for index, atom in enumerate(self.top_layer.motif):
            top_atom_index_dictionary[index] = counter
            counter += self.top_layer.norbitals[int(atom[3])]

        # Add hopping between layers
        h = np.zeros([self.basisdim, self.basisdim])
        for initial_atom, final_atom, _, _ in self.interlayer_bonds:
            final_atom -= self.bottom_layer.natoms
            initial_species = int(self.bottom_layer.motif[initial_atom][3])
            final_species   = int(self.top_layer.motif[final_atom][3])
            i = bottom_atom_index_dictionary[initial_atom]
            i_orbitals = self.bottom_layer.norbitals[initial_species]
            f = top_atom_index_dictionary[final_atom] + self.bottom_layer.basisdim
            f_orbitals = self.top_layer.norbitals[final_species]
            h[i : i + i_orbitals, f : f + f_orbitals] = self.interlayer_hopping*np.eye(i_orbitals, f_orbitals)
        
        h += h.T
        self.hamiltonian[0] += h

        # Update bonds in top_layer to match self.motif and glue all
        top_layer_bonds = []
        for initial_atom, final_atom, cell, nn in self.top_layer.bonds:
            new_bond = [initial_atom + self.bottom_layer.natoms, 
                        final_atom + self.bottom_layer.natoms,
                        cell, nn]
            top_layer_bonds.append(new_bond)

        self.bonds = self.bottom_layer.bonds + top_layer_bonds + self.interlayer_bonds


    def hamiltonian_k(self, k):
        hamiltonian_k = np.zeros([self.basisdim, self.basisdim], dtype=np.complex_)
        for cell_index, cell in enumerate(self._unit_cell_list):
            hamiltonian_k += self.hamiltonian[cell_index] * cmath.exp(1j*np.dot(k, cell))

        return hamiltonian_k


def bethe_lattice(z: int = 3, depth: int = 3, length: float = 1) -> Tuple[list, list]:
    """ 
    Routine to generate a Bethe lattice.
    
    :param z: Coordination number. Defaults to 3
    :param depth: Number of shells of the lattice (central atom is depth 0)
    :param length: Float to specify length of bonds
    :return: Motif: list of all atoms' positions 
    """

    motif_previous_shell = [[0., 0., 0., 0]]
    motif = [[0., 0., 0., 0]]
    cell = [0., 0., 0.]
    bonds = []
    previous_angles = [0]
    bond_length = length
    atom_index = 1
    for i in range(1, depth + 1):
        natoms_in_shell = z*(z - 1)**(i - 1)
        angles = np.linspace(0, 2*np.pi, natoms_in_shell + 1)
        angle_between = angles[1] - angles[0]
        angles += angle_between/2 + previous_angles[0]
        for angle in angles[:-1]:
            atom = bond_length*np.array([np.cos(angle), np.sin(angle), 0, 0])
            motif.append(atom)
            distance = np.linalg.norm(atom[:3] - np.array(motif_previous_shell)[:, :3], axis=1)
            neighbour = np.where(distance == np.min(distance))[0][0]
            bonds.append([atom_index, neighbour, cell, 1])
            atom_index += 1

        motif_previous_shell = np.copy(motif)
        bond_length += length
        previous_angles = angles

    return motif, bonds









from tightbinder.models import SlaterKoster
from tightbinder.disorder import amorphize, introduce_vacancies, introduce_impurities
from tightbinder.fileparse import parse_config_file
from tightbinder.utils import hash_numpy_array
import numpy as np
import math


"""
Tests to check functionality of the disorder module.
"""


def test_amorphize():
    """
    Tests that amorphize can be applied to a model.
    """
    
    np.random.seed(1)
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).supercell(n1=2, n2=2)
    model = amorphize(model, 0.2)
    
    motif_hash = hash_numpy_array(model.motif)
        
    assert math.isclose(motif_hash, 322.32157833573854)
        

def test_vacancies():
    """
    Tests that introduce_vacancies can be applied to a model.
    """
    
    np.random.seed(1)
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).supercell(n1=2, n2=2)
    model = introduce_vacancies(model, 0.2)
    
    motif_hash = hash_numpy_array(model.motif)
        
    assert math.isclose(motif_hash, 171.9359264901)
    

def test_impurities():
    """
    Tests that introduce_impurities can be applied to a model.
    """
    
    np.random.seed(1)
    
    file = open("./examples/hBN.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).supercell(n1=2)
    model.initialize_hamiltonian()
    
    model = introduce_impurities(model, 10, 0.9)
    
    hamiltonian_hash = np.real(hash_numpy_array(model.hamiltonian[0]))
        
    assert math.isclose(hamiltonian_hash, 431.73999999999995)
    
    
    
    


from tightbinder.topology import calculate_wannier_centre_flow, calculate_z2_invariant, calculate_chern_number, chern_marker, specify_partition_plane, entanglement_spectrum, plot_wannier_centre_flow
from tightbinder.models import SlaterKoster, AgarwalaChern
from tightbinder.fileparse import parse_config_file
from tightbinder.utils import hash_numpy_array
import numpy as np
import matplotlib.pyplot as plt
import math

"""
Module with tests to check the topology module.
"""

def test_wannier_centre_flow():
    """
    Tests the calculation of the wannier centre flow for a SlaterKoster model.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 5, refine_mesh=False)
    
    wcc_hash = hash_numpy_array(wcc)
        
    assert math.isclose(wcc_hash, 9.241406000000001)
    

def test_wannier_centre_flow_refined():
    """
    Tests the calculation of the wannier centre flow but with a refined mesh.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 5)
    
    wcc_hash = hash_numpy_array(wcc)
    
    assert math.isclose(wcc_hash, 22.862526000000003)
            

def test_z2_invariant():
    """
    Tests the z2 invariant calculation algorithm.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 10, refine_mesh=False)
    z2 = calculate_z2_invariant(wcc)
    
    assert z2 == 1
    

def test_z2_phase_diagram():
    """
    Tests that the z2 invariant calculation algorithm works to determine a phase diagram, changing the spin-orbit coupling.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 10, refine_mesh=False)
    z2 = calculate_z2_invariant(wcc)
    
    assert z2 == 1
    
    model.configuration["Spin-orbit coupling"][0] = 0.5
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 10, refine_mesh=False)
    z2 = calculate_z2_invariant(wcc)
    
    assert z2 == 0
    
    
def test_chern_wcc():
    """
    Tests calculation of Wannier centers for a Chern insulator.
    """
    
    model = AgarwalaChern(m=-1)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 10, full_BZ=True, refine_mesh=False)
    
    wcc_hash = hash_numpy_array(wcc)
    
    assert math.isclose(wcc_hash,  2.62954)
            

def test_chern_number():
    """
    Tests calculation of Chern number.
    """
    
    model = AgarwalaChern(m=-1)
    model.initialize_hamiltonian()
    
    wcc = calculate_wannier_centre_flow(model, 10, full_BZ=True)
    chern = calculate_chern_number(wcc)
    
    assert chern == 0.9804999999999999
    

def test_chern_marker():
    """
    Tests calculation of Chern marker.
    """
    
    model = AgarwalaChern(m=-1).reduce(n1=10, n2=10)
    model.boundary = "OBC"
    model.initialize_hamiltonian()
    
    results = model.solve()
    
    marker = chern_marker(model, results).reshape(-1, 2)
    marker = np.sum(marker, axis=1)
    
    marker_hash = hash_numpy_array(marker)
        
    assert math.isclose(marker_hash, 197.6131102153834)
        

def test_partition_plane():
    """
    Tests generation of two partitions of the system according to a given plane.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).reduce(n1=10)
    model.initialize_hamiltonian()
    
    plane = [0, 1, 0, np.max(model.motif[:, 1]/2)]
    partition = specify_partition_plane(model, plane)
        
    assert np.allclose(partition, [0, 1, 2, 3, 4, 5, 6, 7, 8, 9])
    

def test_entanglement_spectrum():
    """
    Tests calculation of the entanglement spectrum from a spatial partition of a system.
    """
    
    file = open("./examples/Bi111.txt", "r")
    config = parse_config_file(file)
    
    model = SlaterKoster(config).reduce(n1=5)
    model.initialize_hamiltonian()
    
    plane = [0, 1, 0, np.max(model.motif[:, 1]/2)]
    partition = specify_partition_plane(model, plane)
    entanglement = entanglement_spectrum(model, partition)
    
    entanglement_hash = hash_numpy_array(entanglement)
    
    assert math.isclose(entanglement_hash, 19.351744089214876)
            
    


from tightbinder.models import WilsonAmorphous, HaldaneModel, BHZ, AgarwalaChern
from tightbinder.result import State
from tightbinder.utils import hash_numpy_array
import numpy as np
import matplotlib.pyplot as plt
import math

"""
This scripts contains tests to verify the functionality of the predefined models.
"""

def test_wilson_amorphous():
    """
    Tests that the WilsonAmorphous model can be initialized and obtain the band structure correctly.
    """
    
    model = WilsonAmorphous(m=4, r=1.4)
    model.initialize_hamiltonian()
    
    labels = ["G", "X", "M", "G"]
    kpoints = model.high_symmetry_path(10, labels)
    results = model.solve(kpoints)
    
    energies_hash = hash_numpy_array(results.eigen_energy)
    
    assert math.isclose(energies_hash, 856.0)
            

def test_haldane_model():
    """
    Tests that the HaldaneModel model can be initialized and obtain the band structure correctly.
    """
    
    model = HaldaneModel()
    model.initialize_hamiltonian()
    
    labels = ["G", "M", "K", "G"]
    kpoints = model.high_symmetry_path(10, labels)
    results = model.solve(kpoints)
    
    energies_hash = hash_numpy_array(results.eigen_energy)
        
    assert math.isclose(energies_hash, 76.56623716571589)
    

def test_bhz():
    """
    Tests that the BHZ model can be initialized and obtain the band structure correctly.
    """
    
    model = BHZ(g=1, u=0.0, c=0.0)
    model.initialize_hamiltonian()
    
    labels = ["G", "K", "M", "G"]
    kpoints = model.high_symmetry_path(10, labels)
    results = model.solve(kpoints)
    
    energies_hash = hash_numpy_array(results.eigen_energy)
        
    assert math.isclose(energies_hash, 304.0)

def test_agarwala_chern():
    """
    Tests that the AgarwalaChern model can be initialized and obtain the band structure correctly.
    """
    
    model = AgarwalaChern()
    model.initialize_hamiltonian()
    
    labels = ["G", "K", "M", "G"]
    kpoints = model.high_symmetry_path(10, labels)
    results = model.solve(kpoints)
    
    energies_hash = hash_numpy_array(results.eigen_energy)
        
    assert math.isclose(energies_hash, 206.0)
    

def test_edge_state_amplitude():
    """
    Checks that the amplitude of the edge state is correctly computed.
    """
    
    # Declaration of parameters of the model
    m = 4
    cutoff = 1.4
    disorder = 0.2
    ncells = 12

    # Init. model, create supercell with OBC and amorphize
    model = WilsonAmorphous(m=m, r=cutoff).reduce(n1=ncells, n2=ncells)

    # Obtain the eigenstates of the system
    model.initialize_hamiltonian()
    results = model.solve()

    # Identify a edge state (zero mode) of the system
    state_index = np.argmin(np.abs(results.eigen_energy))
    edge_state = State(results.eigen_states[0][:, state_index], model)
            
    assert np.allclose(np.sort(edge_state.amplitude)[-5:], 
                       [0.01730692, 0.01730692, 0.01730692, 0.01730692, 0.01730692])